import chromadb
from chromadb.config import Settings
import pandas as pd
class Memory:
    def __init__(self, lm, collection_name="escargot_memory"):
        # Initialize ChromaDB client and specify the collection for storing vectors
        self.client = chromadb.PersistentClient(path="./escargot_memory", settings=Settings(allow_reset=True))
        self.collection_name = collection_name
        self.lm = lm

        self.collection = self.client.get_or_create_collection(collection_name)

    def reset_collection(self):
        self.client.reset() 

    def store_memory(self, text, metadata={None:None},collection_name = None):
        # Embed the text using the lm's embed function
        vector = self.lm.get_embedding(text)

        # Add the embedded vector to the collection
        if collection_name is None:
            self.collection.add(ids=text, embeddings=[vector], metadatas=[metadata])
        else:
            collection = self.client.get_collection(collection_name)
            collection.add(ids=text, embeddings=[vector], metadatas=[metadata])

    def query_collection(self,query, max_results=10, collection_name = "escargot_memory", metadata = None):
        collection = self.client.get_collection(collection_name)
        query_embeddings = self.lm.get_embedding(query)
        if metadata is not None:
            results = collection.query(
                query_embeddings=query_embeddings,
                n_results=max_results,
                where=metadata,
                include=['distances']
            )
        else:
            results = collection.query(
                query_embeddings=query_embeddings,
                n_results=max_results,
                include=['distances']
            )
        return results

    def get_all_vectors(self):
        # Return all vectors stored in the collection
        return self.collection.get()

    def delete_vector(self, text):
        # Delete a vector by id (text)
        self.collection.delete(ids=text)



config = {
    "azuregpt35-16k" : {
        "model_id":"gpt-35-turbo-16k", 
        "prompt_token_cost": 0.001,
        "response_token_cost": 0.002,
        "temperature": 0.7,
        "max_tokens": 2000,
        "stop": None,
        "api_version": "",
        "api_base": "",
        "api_key": "",
        "embedding_id":"text-embedding-ada-002"
    },
    "memgraph" : {
        "host": "",
        "port": 7687
    },
    "weaviate" : {
        "api_key": "",
        "url": "",
        "db": "",
        "limit": 200
    }
}

from escargot import Escargot
escargot = Escargot(config, node_types = "BiologicalProcess, BodyPart, CellularComponent, Datatype, Disease, Drug, DrugClass, Gene, MolecularFunction, Pathway, Symptom", relationship_types = """CHEMICALBINDSGENE
CHEMICALDECREASESEXPRESSION
CHEMICALINCREASESEXPRESSION
DRUGINCLASS
DRUGCAUSESEFFECT
DRUGTREATSDISEASE
GENEPARTICIPATESINBIOLOGICALPROCESS
GENEINPATHWAY
GENEINTERACTSWITHGENE
GENEHASMOLECULARFUNCTION
GENEASSOCIATEDWITHCELLULARCOMPONENT
GENEASSOCIATESWITHDISEASE
SYMPTOMMANIFESTATIONOFDISEASE
BODYPARTUNDEREXPRESSESGENE
BODYPARTOVEREXPRESSESGENE
DISEASELOCALIZESTOANATOMY
DISEASEASSOCIATESWITHDISEASET""", 
model_name="azuregpt35-16k")
escargot.memgraph_client.schema = """Node properties are the following:
Node name: 'BiologicalProcess', Node properties: ['commonName']
Node name: 'BodyPart', Node properties: ['commonName']
Node name: 'CellularComponent', Node properties: ['commonName']
Node name: 'Disease', Node properties: ['commonName']
Node name: 'Drug', Node properties: ['commonName']
Node name: 'DrugClass', Node properties: ['commonName']
Node name: 'Gene', Node properties: ['commonName', 'geneSymbol', 'typeOfGene']
Node name: 'MolecularFunction', Node properties: ['commonName']
Node name: 'Pathway', Node properties: ['commonName']
Node name: 'Symptom', Node properties: ['commonName']
Relationship properties are the following:
The relationships are the following:
(:Drug)-[:CHEMICALBINDSGENE]-(:Gene)
(:Drug)-[:CHEMICALDECREASESEXPRESSION]-(:Gene)
(:Drug)-[:CHEMICALINCREASESEXPRESSION]-(:Gene)
(:Drug)-[:DRUGINCLASS]-(:DrugClass)
(:Drug)-[:DRUGCAUSESEFFECT]-(:Disease)
(:Drug)-[:DRUGTREATSDISEASE]-(:Disease)
(:Gene)-[:GENEPARTICIPATESINBIOLOGICALPROCESS]-(:BiologicalProcess)
(:Gene)-[:GENEINPATHWAY]-(:Pathway)
(:Gene)-[:GENEINTERACTSWITHGENE]-(:Gene)
(:Gene)-[:GENEHASMOLECULARFUNCTION]-(:MolecularFunction)
(:Gene)-[:GENEASSOCIATEDWITHCELLULARCOMPONENT]-(:CellularComponent)
(:Gene)-[:GENEASSOCIATESWITHDISEASE]-(:Disease)
(:Symptom)-[:SYMPTOMMANIFESTATIONOFDISEASE]-(:Disease)
(:BodyPart)-[:BODYPARTUNDEREXPRESSESGENE]-(:Gene)
(:BodyPart)-[:BODYPARTOVEREXPRESSESGENE]-(:Gene)
(:Disease)-[:DISEASELOCALIZESTOANATOMY]-(:BodyPart)
(:Disease)-[:DISEASEASSOCIATESWITHDISEASET]-(:Disease)"""


import json
import dill
json_files =  ['MCQ_1hop.json', 'MCQ_2hop.json', 'OpenEnded_1hop.json', 'OpenEnded_2hop.json', 'True_or_False_1hop.json', 'True_or_False_2hop.json']
responses = {}
for json_file in json_files:
    print(json_file)
    with open("../dataset/"+json_file) as f:
        data = json.load(f)
    responses[json_file] = {}
    for question in data:
        response = ''
        tries = 0
        while response == '' and tries < 3:
            escargot.memgraph_client.cache = {}
            try:
                response = escargot.ask(question['question'], answer_type= "array",debug_level = 0)
            except Exception as e:
                response = ''
            tries += 1
        
        print('question:', question['question'], 'answer:', question['answer'], 'response:', response)
        print("------------------------------------------------------------------------------------------------------------------------------\n")
        responses[json_file][question['question']] = str(response)
dill.dump(responses, open('Escargot_esponses.pkl', 'wb'))


from openai import AzureOpenAI

def get_embedding(api_key,api_base,api_version,deployment_name, text_to_embed):
    client = AzureOpenAI(
        api_key=api_key,
        api_version=api_version,
        azure_endpoint = api_base
    )
    response = client.embeddings.create(
        model=deployment_name,
        input=text_to_embed
    )

    return response.data[0].embedding


from gqlalchemy import Memgraph
#from langchain_community.graphs import MemgraphGraph
import os
from typing import Dict
import json
import logging

class MemgraphClient:
    def __init__(self, config_path, logger):
        self.logger = logger
        self.config: Dict = None
        if type(config_path) == dict:
            self.config = config_path
        else:
            self.load_config(config_path)
        if "memgraph" not in self.config:
            self.memgraph = None
        else:
            self.config: Dict = self.config["memgraph"]
            self.host = self.config["host"]
            self.port = self.config["port"]
            self.memgraph = Memgraph(host=self.host, port=self.port)
            self.num_responses = 7
            self.cache = {}
            self.schema = None

    def get_schema(self):
        SCHEMA_QUERY = """
        CALL llm_util.schema("raw")
        YIELD *
        RETURN *
        """
        tries = 3
        while tries > 0:
            try:
                db_structured_schema = list(self.memgraph.execute_and_fetch(SCHEMA_QUERY))
                assert db_structured_schema is not None
                structured_schema = db_structured_schema[0]['schema']
                break
            except Exception as e:
                tries -= 1
                self.logger.error(f"Error in memgraph: {e}, trying again {tries}")
        
        formatted_node_props = []
        node_names = []

        for node_name, properties in structured_schema["node_props"].items():
            node_names.append(node_name)
            formatted_node_props.append(
                f"Node name: '{node_name}', Node properties: {properties}"
            )

        formatted_rels = []
        relationship_names = []
        for rel in structured_schema["relationships"]:
            relationship_names.append(rel['type'])
            formatted_rels.append(
                f"(:{rel['start']})-[:{rel['type']}]->(:{rel['end']})"
            )

        self.schema = "\n".join(
            [
                "Node properties are the following:",
                *formatted_node_props,
                "The relationships are the following:",
                *formatted_rels,
            ]
        )
        return (", ").join(node_names), ("\n").join(relationship_names)

    def load_config(self, path: str) -> None:
        """
        Load configuration from a specified path.

        :param path: Path to the config file. If an empty path provided,
                     default is `config.json` in the current directory.
        :type path: str
        """
        if path == "":
            current_dir = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(current_dir, "config.json")

        with open(path, "r") as f:
            self.config = json.load(f)

    def clear_cache(self) -> None:
        """
        Clear the response cache.
        """
        self.cache = {}

    def execute(self, lm, query, statement):
        # if statement in self.cache:
        #     results = self.cache[statement][1]
        # else:
            num_responses = self.num_responses
            response = ''
            responses_hash = set()
            memgraph_results = []
            iter = 0
            responses = lm.get_response_texts(
                lm.query(query, num_responses=num_responses)
            )
            empty_responses = True
            while iter < num_responses and empty_responses:
                try:
                    response = responses[iter]
                    iter += 1
                    # Check if the response is in the hash set
                    if response in responses_hash:
                        continue
                    responses_hash.add(response)
                    # Remove "Answer:" from the response
                    if response.startswith("Answer:"):
                        response= response[8:].strip()
                    
                    if response == "" or response == None:
                        continue
                    #remove ```cypher from the response
                    response = response.replace("```cypher", "")

                    #remove ``` from anywhere in the response
                    response = response.replace("```", "")

                    #remove \n from the response
                    response = response.replace("\n", " ")

                    #TODO: tweak if necessary. Get rid of directionality in the response
                    response = response.replace("<-", "-")
                    response = response.replace("->", "-")
                    
                    self.logger.info(f"Executing memgraph for statement: {statement}, response: {response}")
                    memgraph_results = self.memgraph.execute_and_fetch(response)
                    memgraph_results = list(memgraph_results)
                    if memgraph_results != []:
                        empty_responses = False
                except Exception as e:
                    self.logger.error(f"Error in memgraph: {e}, trying again {iter}")
            self.logger.info(f"Memgraph results: {memgraph_results}")
            results = []
            # get the value in the dictionary x in memgraph_results
            for value in memgraph_results:
                for key, val in value.items():
                    if val != [] and val != None and val != "":
                        if type(val) == list:
                            for v in val:
                                results.append(v)
                        else:
                            results.append(val)
            #get unique values
            results = list(set(results))
            # self.cache[statement] = [response,results]

            return results


# Copyright (c) 2023 ETH Zurich.
#                    All rights reserved.
#
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
#
# main author: Nils Blach

import backoff
import os
import random
import time
from typing import List, Dict, Union
from openai import OpenAI, OpenAIError
from openai.types.chat.chat_completion import ChatCompletion

from .abstract_language_model import AbstractLanguageModel


class ChatGPT(AbstractLanguageModel):
    """
    The ChatGPT class handles interactions with the OpenAI models using the provided configuration.

    Inherits from the AbstractLanguageModel and implements its abstract methods.
    """

    def __init__(
        self, config_path: str = "", model_name: str = "chatgpt", cache: bool = False
    ) -> None:
        """
        Initialize the ChatGPT instance with configuration, model details, and caching options.

        :param config_path: Path to the configuration file. Defaults to "".
        :type config_path: str
        :param model_name: Name of the model, default is 'chatgpt'. Used to select the correct configuration.
        :type model_name: str
        :param cache: Flag to determine whether to cache responses. Defaults to False.
        :type cache: bool
        """
        super().__init__(config_path, model_name, cache)
        self.config: Dict = self.config[model_name]
        # The model_id is the id of the model that is used for chatgpt, i.e. gpt-4, gpt-3.5-turbo, etc.
        self.model_id: str = self.config["model_id"]
        # The prompt_token_cost and response_token_cost are the costs for 1000 prompt tokens and 1000 response tokens respectively.
        self.prompt_token_cost: float = self.config["prompt_token_cost"]
        self.response_token_cost: float = self.config["response_token_cost"]
        # The temperature of a model is defined as the randomness of the model's output.
        self.temperature: float = self.config["temperature"]
        # The maximum number of tokens to generate in the chat completion.
        self.max_tokens: int = self.config["max_tokens"]
        # The stop sequence is a sequence of tokens that the model will stop generating at (it will not generate the stop sequence).
        self.stop: Union[str, List[str]] = self.config["stop"]
        # The account organization is the organization that is used for chatgpt.
        self.organization: str = self.config["organization"]
        if self.organization == "":
            self.logger.warning("OPENAI_ORGANIZATION is not set")
        self.api_key: str = os.getenv("OPENAI_API_KEY", self.config["api_key"])
        if self.api_key == "":
            raise ValueError("OPENAI_API_KEY is not set")
        # Initialize the OpenAI Client
        self.client = OpenAI(api_key=self.api_key, organization=self.organization)

    def query(
        self, query: str, num_responses: int = 1
    ) -> Union[List[ChatCompletion], ChatCompletion]:
        """
        Query the OpenAI model for responses.

        :param query: The query to be posed to the language model.
        :type query: str
        :param num_responses: Number of desired responses, default is 1.
        :type num_responses: int
        :return: Response(s) from the OpenAI model.
        :rtype: Dict
        """
        if self.cache and query in self.respone_cache:
            return self.respone_cache[query]

        if num_responses == 1:
            response = self.chat([{"role": "user", "content": query}], num_responses)
        else:
            response = []
            next_try = num_responses
            total_num_attempts = num_responses
            while num_responses > 0 and total_num_attempts > 0:
                try:
                    assert next_try > 0
                    res = self.chat([{"role": "user", "content": query}], next_try)
                    response.append(res)
                    num_responses -= next_try
                    next_try = min(num_responses, next_try)
                except Exception as e:
                    next_try = (next_try + 1) // 2
                    self.logger.warning(
                        f"Error in chatgpt: {e}, trying again with {next_try} samples"
                    )
                    time.sleep(random.randint(1, 3))
                    total_num_attempts -= 1

        if self.cache:
            self.respone_cache[query] = response
        return response

    @backoff.on_exception(backoff.expo, OpenAIError, max_time=10, max_tries=6)
    def chat(self, messages: List[Dict], num_responses: int = 1) -> ChatCompletion:
        """
        Send chat messages to the OpenAI model and retrieves the model's response.
        Implements backoff on OpenAI error.

        :param messages: A list of message dictionaries for the chat.
        :type messages: List[Dict]
        :param num_responses: Number of desired responses, default is 1.
        :type num_responses: int
        :return: The OpenAI model's response.
        :rtype: ChatCompletion
        """
        response = self.client.chat.completions.create(
            model=self.model_id,
            messages=messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            n=num_responses,
            stop=self.stop,
        )

        self.prompt_tokens += response.usage.prompt_tokens
        self.completion_tokens += response.usage.completion_tokens
        prompt_tokens_k = float(self.prompt_tokens) / 1000.0
        completion_tokens_k = float(self.completion_tokens) / 1000.0
        self.cost = (
            self.prompt_token_cost * prompt_tokens_k
            + self.response_token_cost * completion_tokens_k
        )
        return response

    def get_response_texts(
        self, query_response: Union[List[ChatCompletion], ChatCompletion]
    ) -> List[str]:
        """
        Extract the response texts from the query response.

        :param query_response: The response dictionary (or list of dictionaries) from the OpenAI model.
        :type query_response: Union[List[ChatCompletion], ChatCompletion]
        :return: List of response strings.
        :rtype: List[str]
        """
        if not isinstance(query_response, List):
            query_response = [query_response]
        return [
            choice.message.content
            for response in query_response
            for choice in response.choices
        ]



import openai

def get_embedding(text_to_embed, config_RAG_TEST):

    # Azure 
    openai.api_type ="azure"
    openai.api_key = config_RAG_TEST["azuregpt35-16k"]["api_key"]

    openai.api_version = "2023-05-15"

    openai.azure_endpoint = "https://caire-azure-openai.openai.azure.com/"


    embedding_id = config_RAG_TEST["azuregpt35-16k"]["embedding_id"]

    response = openai.embeddings.create(
        model=embedding_id,
        input=text_to_embed
    )

    embedding = response.data[0].embedding

    return embedding


from __future__ import annotations
import logging
from enum import Enum
from typing import List, Iterator, Dict, Callable, Union
from abc import ABC, abstractmethod
import itertools
import copy
import re
from .utils import remove_quotes, process_knowledge_ids, apply_function

from escargot.operations.thought import Thought
from escargot.language_models import AbstractLanguageModel
from escargot.prompter import ESCARGOTPrompter
from escargot.parser import ESCARGOTParser
from escargot.coder import Coder

class OperationType(Enum):
    """
    Enum to represent different operation types that can be used as unique identifiers.
    """

    generate: int = 0


class Operation(ABC):
    """
    Abstract base class that defines the interface for all operations.
    """

    _ids: Iterator[int] = itertools.count(0)

    operation_type: OperationType = None

    def __init__(self) -> None:
        """
        Initializes a new Operation instance with a unique id, and empty predecessors and successors.
        """
        self.logger: logging.Logger = None
        self.id: int = next(Operation._ids)
        self.predecessors: List[Operation] = []
        self.successors: List[Operation] = []
        self.executed: bool = False
        self.coder = None

    def can_be_executed(self) -> bool:
        """
        Checks if the operation can be executed based on its predecessors.

        :return: True if all predecessors have been executed, False otherwise.
        :rtype: bool
        """
        return all(predecessor.executed for predecessor in self.predecessors)

    def get_previous_thoughts(self) -> List[Thought]:
        """
        Iterates over all predecessors and aggregates their thoughts.

        :return: A list of all thoughts from the predecessors.
        :rtype: List[Thought]
        """
        previous_thoughts: List[Thought] = [
            thought
            for predecessor in self.predecessors
            for thought in predecessor.get_thoughts()
        ]

        return previous_thoughts

    def add_predecessor(self, operation: Operation) -> None:
        """
        Add a preceding operation and update the relationships.

        :param operation: The operation to be set as a predecessor.
        :type operation: Operation
        """
        self.predecessors.append(operation)
        operation.successors.append(self)

    def add_successor(self, operation: Operation) -> None:
        """
        Add a succeeding operation and update the relationships.

        :param operation: The operation to be set as a successor.
        :type operation: Operation
        """
        self.successors.append(operation)
        operation.predecessors.append(self)

    def execute(
        self, lm: AbstractLanguageModel, prompter: ESCARGOTPrompter, parser: ESCARGOTParser, got_steps: Dict, logger : logging.Logger, coder : Coder, **kwargs
    ) -> None:
        """
        Execute the operation, assuring that all predecessors have been executed.

        :param lm: The language model to be used.
        :type lm: AbstractLanguageModel
        :param prompter: The prompter for crafting prompts.
        :type prompter: ESCARGOTPrompter
        :param parser: The parser for parsing responses.
        :type parser: ESCARGOTParser
        :param got_steps: The dictionary of steps in the Graph of Operations.
        :type got_steps: Dict
        :param kwargs: Additional parameters for execution.
        """
        assert self.can_be_executed(), "Not all predecessors have been executed"
        self.logger = logger
        self.coder = coder
        self.logger.debug(
            "Beginning execution of operation %d of type %s", self.id, self.operation_type
        )
        self._execute(lm, prompter, parser, got_steps, **kwargs)
        self.logger.debug("Operation %d executed", self.id)
        self.executed = True

    @abstractmethod
    def _execute(
        self, lm: AbstractLanguageModel, prompter: ESCARGOTPrompter, parser: ESCARGOTParser, got_steps: Dict, **kwargs
    ) -> None:
        """
        Abstract method for the actual execution of the operation.
        This should be implemented in derived classes.

        :param lm: The language model to be used.
        :type lm: AbstractLanguageModel
        :param prompter: The prompter for crafting prompts.
        :type prompter: ESCARGOTPrompter
        :param parser: The parser for parsing responses.
        :type parser: ESCARGOTParser
        :param kwargs: Additional parameters for execution.
        """
        pass

    @abstractmethod
    def get_thoughts(self) -> List[Thought]:
        """
        Abstract method to retrieve the thoughts associated with the operation.
        This should be implemented in derived classes.

        :return: List of associated thoughts.
        :rtype: List[Thought]
        """
        pass


class Generate(Operation):
    """
    Operation to generate thoughts.
    """

    operation_type: OperationType = OperationType.generate

    def __init__(
        self, num_branches_prompt: int = 1, num_branches_response: int = 1
    ) -> None:
        """
        Initializes a new Generate operation.

        :param num_branches_prompt: Number of responses that each prompt should generate (passed to prompter). Defaults to 1.
        :type num_branches_prompt: int
        :param num_branches_response: Number of responses the LM should generate for each prompt. Defaults to 1.
        :type num_branches_response: int
        """
        super().__init__()
        self.num_branches_prompt: int = num_branches_prompt
        self.num_branches_response: int = num_branches_response
        self.thoughts: List[Thought] = []

    def get_thoughts(self) -> List[Thought]:
        """
        Returns the thoughts associated with the operation.

        :return: List of generated thoughts.
        :rtype: List[Thought]
        """
        return self.thoughts
    
    def generate_from_single_thought(
        self, lm: AbstractLanguageModel, prompter: ESCARGOTPrompter, parser: ESCARGOTParser, base_state: Dict
    ):
        prompts = []
        responses = []
        new_states = []
        if len(self.thoughts) > 0 and self.thoughts[-1].state["phase"] == "output":
            temp_state = copy.deepcopy(base_state)
            temp_state["phase"] = "output"
            temp_state['input'] = self.coder.step_output
            prompts.append(prompter.generate_prompt( **temp_state))
            self.logger.debug("Prompt for LM: \n%s", prompts[-1])
        elif "StepID" in base_state and "instruction" in base_state and base_state["instruction"]["Code"] is not None:
            code = base_state["instruction"]["Code"][0]
            new_code, compiled = self.coder.execute_code(code, base_state["instruction"]["Instruction"], base_state["StepID"], prompter, self.logger)
            new_state  = {**base_state, "input": new_code, "compiled": compiled}
            new_state["instructions"][int(base_state["StepID"])-1]['Code'] = [new_code]
            new_states.append(new_state)
            self.logger.info("Code: %s", new_code)
            self.logger.info("Compiled: %s", compiled)
            
            return prompts, responses, new_states
        else:
            prompts.append(prompter.generate_prompt( **base_state))
        for prompt in prompts:
            if prompt is None or prompt == "":
                self.logger.warning("Prompt for LM is empty")
                return prompt, []
            self.logger.debug("Prompt for LM: %s", prompt)
            
            tries = 0
            while tries < 3:
                try:
                    new_states = []
                    responses = []
                    lm_responses =lm.get_response_texts(
                        lm.query(prompt, num_responses=self.num_branches_response)
                    )
                    for response in lm_responses:
                        responses.append(response)
                        new_state = parser.parse_generate_answer(base_state, response)
                        new_states.append(new_state)
                        self.logger.debug("Response from LM: %s", response)
                    break
                except Exception as e:
                    self.logger.warning("Error in LM: %s, trying again with prompt: %s", e, prompt)
                    tries += 1

        if new_states != [] and "select_highest_score" in new_states[0]:
            #empty list of length len(new_states)
            scores = [new_state["scores"] for new_state in new_states]
            #element wise sum of the scores
            sum_scores = [sum(elem) for elem in zip(*scores)]
            #get the index of the highest score
            highest_score_index = sum_scores.index(max(sum_scores))
            responses = [sum_scores]
            new_states = [new_states[0]]
            new_states[0]["input"] = base_state["input"][highest_score_index]
            new_states[0]["scores"] = sum_scores
            if new_states[0]["phase"] == "python_conversion":
                new_states[0]["full_plan"] = new_states[0]["input"]
                self.logger.warning("Strategy:\n%s", new_states[0]["input"])
            new_states[0].pop("select_highest_score")
            
        
        return prompts, responses, new_states

    def _execute(
        self, lm: AbstractLanguageModel, prompter: ESCARGOTPrompter, parser: ESCARGOTParser, got_steps: Dict, **kwargs
    ) -> None:
        """
        Executes the Generate operation by generating thoughts from the predecessors.
        The thoughts are generated by prompting the LM with the predecessors' thought states.
        If there are no predecessors, the kwargs are used as a base state.

        :param lm: The language model to be used.
        :type lm: AbstractLanguageModel
        :param prompter: The prompter for crafting prompts.
        :type prompter: ESCARGOTPrompter
        :param parser: The parser for parsing responses.
        :type parser: ESCARGOTParser
        :param got_steps: The dictionary of steps in the Graph of Operations.
        :type got_steps: Dict
        :param kwargs: Additional parameters for execution.
        """
        previous_thoughts: List[Thought] = self.get_previous_thoughts()

        if len(previous_thoughts) == 0 and len(self.predecessors) > 0:
            return

        if len(previous_thoughts) == 0:
            # no predecessors, use kwargs as base state
            previous_thoughts = [Thought(state=kwargs)]

        if len(previous_thoughts) == 1:
        # for thought in previous_thoughts:
            base_state = previous_thoughts[0].state
            if "generate_successors" in base_state:
                base_state.pop("generate_successors")
            if "num_branches_response" in base_state:
                self.num_branches_response = base_state.pop("num_branches_response")
            
        elif len(previous_thoughts) > 1:

            base_states = [thought.state for thought in previous_thoughts]
            input_array = {}
            for base_state in base_states:
                input_array[base_state["StepID"]] = base_state["input"]
            base_state = copy.deepcopy(base_states[0])
            base_state["input"] = input_array
            if "instruction" in base_state:
                base_state.pop("instruction")
        if len(self.get_thoughts()) > 0 and "StepID" in self.get_thoughts()[0].state:
            base_state["StepID"] = self.get_thoughts()[0].state["StepID"]
            for instruction in base_state["instructions"]:
                if instruction["StepID"] == base_state["StepID"] or instruction["StepID"] == int(base_state["StepID"]) or instruction["StepID"] == str("StepID_" + base_state["StepID"]):
                    base_state["instruction"] = instruction
                    break
            
        prompts, responses, new_states = self.generate_from_single_thought(
            lm, prompter, parser, base_state
        )
        self.logger.info("Responses from LM: %s", responses)
        index = 0
        for new_state in new_states:
            if new_state is not None:
                if len(previous_thoughts) > 1 and len(prompts) >= 1:
                    new_state = {**base_state, **new_state, "prompt": prompts[index]}
                elif len(prompts) >= 1:
                    new_state = {**base_state, **new_state, "prompt": prompts[0]}
                else:
                    new_state = {**base_state, **new_state, "prompt": prompts}
                if len(self.thoughts) > 0:
                    if self.thoughts[-1].state["phase"] == "plan_assessment" or self.thoughts[-1].state["phase"] == "code_assessment":
                        if type(self.thoughts[-1].state["input"]) == str:
                            self.thoughts[-1].state["input"] = [self.thoughts[-1].state["input"]]
                        self.thoughts[-1].state["input"].append(new_state["input"])
                        continue
                    elif self.thoughts[-1].state["phase"] == "output":
                        self.thoughts[-1].state["input"] = new_state["input"]
                        self.thoughts[-1].state["prompt"] = new_state["prompt"]
                    else:
                        self.thoughts[-1].state = new_state
                else:
                    self.thoughts.append(Thought(new_state))

                if "generate_successors" in new_state:
                    predecessor = self
                    gen_nda = Generate(1, new_state["generate_successors"])
                    gen_nda.add_predecessor(predecessor)
                        
                
                # edges look like ['StepID_1-StepID_2', 'StepID_1-StepID_3', 'StepID_2-StepID_4', 'StepID_3-StepID_4']
                # it can also look like ['1-2', '1-3', '2-4', '3-4']
            
                if "edges" in new_state and "StepID" not in new_state:
                    #only one edge
                    if new_state["edges"] == []:
                        got_steps['1'] = Generate(1, 1)
                        got_steps['1'].thoughts = [Thought(
                            state={**self.thoughts[-1].state, "StepID": '1'}
                        )]
                    else:
                        for edge in new_state["edges"]:
                            edge = edge.split('-')
                            #remove "StepID" from each element in edge
                            edge = [elem.replace("StepID_", "") for elem in edge]
                            edge = [elem.replace("StepID", "") for elem in edge]

                            #sort the edge
                            edge.sort()

                            if edge[0] not in got_steps:
                                got_steps[edge[0]] = Generate(1, 1)
                                got_steps[edge[0]].thoughts = [Thought(
                                    state={**self.thoughts[-1].state, "StepID": edge[0]}
                                )]
                            if edge[1] not in got_steps:
                                got_steps[edge[1]] = Generate(1, 1)
                                got_steps[edge[1]].thoughts = [Thought(
                                    state={**self.thoughts[-1].state, "StepID": edge[1]}
                                )]
                            got_steps[edge[0]].add_successor(got_steps[edge[1]])
                        #check for all steps in got_steps for predecessors. If no predecessors, assign the self as predecessor
                    for step in got_steps:
                        if len(got_steps[step].predecessors) == 0:
                            got_steps[step].add_predecessor(self)
                    
                    #get the last integer key in the got_steps
                    got_steps_keys = list(got_steps.keys())
                    got_steps_keys = [int(elem) for elem in got_steps_keys]
                    got_steps_keys.sort()
                    got_steps_keys = [str(elem) for elem in got_steps_keys]
                    last_step = got_steps_keys[-1]
                    got_steps[str(last_step)].add_successor(Generate(1, 1))
                    got_steps[str(last_step)].successors[-1].thoughts = [Thought(
                        state={**self.thoughts[-1].state, "phase": "output"}
                    )]
                    got_steps[str(last_step)].successors[-1].thoughts[-1].state["input"] = ""
                    new_state.pop("edges")
                    
            index += 1

        if self.thoughts[-1].state["phase"] == "output":
            self.logger.info("Output: %s", self.thoughts[-1].state["input"])
        self.logger.debug(
            "Generate operation %d created %d new thoughts", self.id, len(self.thoughts)
        )




def remove_quotes(s):
    if s and s[0] in ("'", '"') and s[0] == s[-1]:
        return s[1:-1]
    return s

def process_knowledge_ids(knowledge_ids):
    knowledge_ids = knowledge_ids[knowledge_ids.find("(")+1:knowledge_ids.find(")")]
    knowledge_ids = [remove_quotes(elem.strip()) for elem in knowledge_ids.split(",")]
    return knowledge_ids

def apply_function(knowledge_ids, func_name, knowledge_list):
    intersect, union, difference = set(), set(), set()
    
    def update_set(op_set, current_list, op):
        if op:
            if not op_set:
                op_set.update(current_list)
            else:
                op_set.update(current_list) if op == 'union' else op_set.difference_update(current_list)

    for knowledge_id in knowledge_ids:
        if isinstance(knowledge_id, list):
            current_set = set(knowledge_id)
        else:
            current_set = set(knowledge_list.get(knowledge_id, []))
        if func_name == 'intersect':
            update_set(intersect, current_set, 'intersect')
        elif func_name == 'union':
            update_set(union, current_set, 'union')
        elif func_name == 'difference':
            update_set(difference, current_set, 'difference')

    return list(intersect or union or difference)

def get_knowledge_list_from_input(new_state, logger):
    #new_state["input"] is a string and needs to be converted to an array. Need to do a try except to catch any errors
    try:
        if new_state["input"] == "[]":
            knowledge_list_array = []
        #check if the array elements has "" or '' around them. If so add them for each elemen
        elif new_state["input"].count("'") >= 2 or new_state["input"].count('"') >= 2:
            knowledge_list_array = new_state["input"].replace("[", "").replace("]", "").replace("\"","").replace("'", "").split(",")
        elif new_state["input"].count("'") == 0 and new_state["input"].count('"') == 0:
            knowledge_list_array = new_state["input"].replace("[", "").replace("]", "").split(",")
        else:
            knowledge_list_array = eval(new_state["input"])
    except Exception as e:
        logger.error(f"Could not convert input to array: {new_state['input']}. Encountered exception: {e}")
        knowledge_list_array = []
    knowledge_list_array = [elem.strip() for elem in knowledge_list_array]
    return knowledge_list_array


#TODO:
#semi-auto, full-auto modes:
    #step function to step into the next on the execution queue
    #chat log function for outputting later
#just like got_steps variable, we need an entire edge list and accessibility to any step
#accessing and rerunning steps (what are the existing values that need to be updated, new variables? [reran, attempted, previous_executions])


import escargot.language_models as language_models
import logging
import io

#vectorized knowledge
from escargot.vector_db.weaviate import WeaviateClient

#memgraph Cypher
import escargot.cypher.memgraph as memgraph

class Escargot:

    def __init__(self, config: str, node_types:str = "", relationship_types:str = "", model_name: str = "azuregpt35-16k"):
        logger = logging.getLogger(__name__)
        self.logger = logger
        self.log = ""
        self.lm = language_models.AzureGPT(config, model_name=model_name, logger=logger)
        self.vdb = WeaviateClient(config, self.logger)
        self.memory = None
        self.node_types = ""
        self.relationship_types = ""
        self.question = ""
        self.controller = None
        self.operations_graph = None
        if self.vdb.client is None:
            self.vdb = None
        self.memgraph_client = memgraph.MemgraphClient(config, logger)
        if self.memgraph_client.memgraph is None:
            self.memgraph_client = None
        else:
            if node_types == "" or relationship_types == "":
                self.node_types, self.relationship_types = self.memgraph_client.get_schema()
            else:
                self.node_types = node_types
                self.relationship_types = relationship_types
    
    def setup_logger(self, debug_level):
        log_stream = io.StringIO()
        f_handler = logging.StreamHandler(log_stream)
        c_handler = logging.StreamHandler()
        if debug_level == 0:
            self.logger.setLevel(logging.ERROR)
            c_handler.setLevel(logging.ERROR)
            f_handler.setLevel(logging.ERROR)
        elif debug_level == 1:
            self.logger.setLevel(logging.WARNING)
            c_handler.setLevel(logging.WARNING)
            f_handler.setLevel(logging.WARNING)
        elif debug_level == 2:
            self.logger.setLevel(logging.INFO)
            c_handler.setLevel(logging.INFO)
            f_handler.setLevel(logging.INFO)
        elif debug_level == 3:
            self.logger.setLevel(logging.DEBUG)
            c_handler.setLevel(logging.DEBUG)
            f_handler.setLevel(logging.DEBUG)
        c_format = logging.Formatter('%(asctime)s - %(filename)s - %(funcName)s(%(lineno)d) - %(message)s')
        c_handler.setFormatter(c_format)
        f_handler.setFormatter(c_format)
        self.logger.addHandler(f_handler)
        self.logger.addHandler(c_handler)

        return log_stream, c_handler, f_handler

    def finalize_logger(self,log_stream, c_handler, f_handler):
        self.log += log_stream.getvalue()
        #reset logger
        self.logger.removeHandler(c_handler)
        c_handler.close()
        self.logger.removeHandler(f_handler)
        f_handler.close()

    #debug_level: 0, 1, 2, 3
    #0: no debug, only output
    #1: output, instructions, and exceptions
    #2: output, instructions, exceptions, and debug info
    #3: output, instructions, exceptions, debug info, and LLM output
    def ask(self, question, answer_type = 'natural', num_strategies=3, debug_level = 0, memory_name = "escargot_memory"):
        """
        Ask a question and get an answer.

        :param question: The question to ask.
        :type question: str
        :param answer_type: The type of answer to expect. Defaults to 'natural'. Options are 'natural', 'array'.
        :type answer_type: str
        :param num_strategies: The number of strategies to generate. Defaults to 3.
        :type num_strategies: int
        :return: The answer to the question.
        :rtype: str
        """
        import escargot.memory as memory
        self.memory = memory.Memory(self.lm, collection_name = memory_name)
        #setup logger
        log_stream, c_handler, f_handler = self.setup_logger(debug_level)
        from escargot import operations
        def got() -> operations.GraphOfOperations:
            operations_graph = operations.GraphOfOperations()

            instruction_node = operations.Generate(1, 1)
            operations_graph.append_operation(instruction_node)
            
            return operations_graph
        
        # Create the Controller
        got = got()
        try:
            from escargot.parser import ESCARGOTParser
            from escargot.prompter import ESCARGOTPrompter
            import escargot.controller as controller
            self.controller = controller.Controller(
                self.lm, 
                got, 
                ESCARGOTPrompter(memgraph_client = self.memgraph_client,vector_db = self.vdb, lm=self.lm,node_types=self.node_types,relationship_types=self.relationship_types, logger = self.logger),
                ESCARGOTParser(self.logger),
                self.logger,
                {
                    "question": question,
                    "input": "",
                    "phase": "planning",
                    "method" : "got",
                    "num_branches_response": num_strategies,
                    "answer_type": answer_type
                }
            )
            self.controller.run()
        except Exception as e:
            self.logger.error("Error executing controller: %s", e)

        self.operations_graph = self.controller.graph.operations
        output = ""
        if self.controller.final_thought is not None:
            if answer_type == 'natural':
                output = self.controller.final_thought.state['input']
            elif answer_type == 'array':
                # output = list(list(self.controller.coder.step_output.values())[-1].values())[-1]
                output = list(self.controller.coder.step_output.values())[-1]

        self.logger.warning(f"Output: {output}")

        #remove logger
        self.finalize_logger(log_stream, c_handler, f_handler)

        return output
    
    def initialize_controller(self, question, answer_type = 'natural', num_strategies=3, debug_level = 0):
        
        #setup logger
        log_stream, c_handler, f_handler = self.setup_logger(debug_level)

        def got() -> operations.GraphOfOperations:
            operations_graph = operations.GraphOfOperations()

            instruction_node = operations.Generate(1, 1)
            operations_graph.append_operation(instruction_node)
            
            return operations_graph
        
        # Create the Controller
        got = got()
        try:
            from escargot.parser import ESCARGOTParser
            from escargot.prompter import ESCARGOTPrompter
            self.controller = controller.Controller(
                self.lm, 
                got, 
                ESCARGOTPrompter(memgraph_client = self.memgraph_client,vector_db = self.vdb, lm=self.lm,node_types=self.node_types,relationship_types=self.relationship_types, logger = self.logger),
                ESCARGOTParser(self.logger),
                self.logger,
                {
                    "question": question,
                    "input": "",
                    "phase": "planning",
                    "method" : "got",
                    "num_branches_response": num_strategies,
                    "answer_type": answer_type
                }
            )
        except Exception as e:
            self.logger.error("Error initializing controller: %s", e)
        self.finalize_logger(log_stream, c_handler, f_handler)

    def step(self):
        #setup logger
        log_stream, c_handler, f_handler = self.setup_logger(self.logger.level)
        try:
            self.controller.execute_step()
        except Exception as e:
            self.logger.error("Error executing controller: %s", e)

        output = ""
        if self.controller.final_thought is not None:
            self.operations_graph = self.controller.graph.operations
            output = self.controller.final_thought.state['input']

        self.logger.warning(f"Output: {output}")

        #remove logger
        self.finalize_logger(log_stream, c_handler, f_handler)
        return output
    
    def quick_chat(self,chat, num_responses=1):
        if self.lm is not None:
            response = self.lm.get_response_texts(
                    self.lm.query(chat, num_responses=num_responses)
                )
            return response[0]
        else:
            return "No language model available."



import json

def output_controller(operations_graph):

    output = []
    # for operation in operations_log['MCQ_1hop.json']['Which of the following binds to the drug Leucovorin? 1. CAD 2. PDS5B 3. SEL1L 4. ABCC2 5. RMI1']:
    # print(len(operations_graph))
    index = 0
    operation = operations_graph[0]
    while len(operation.successors) > 0:
    # for operation in operations_graph:
        
        operation_serialized = {
            "id": "node_"+str(index),
            "operation": operation.operation_type.name,
            "thoughts": [thought.state for thought in operation.get_thoughts()],
        }
        print(operation_serialized["thoughts"][0]["prompt"])
        
        output.append(operation_serialized)
        index = index + 1
        operation = operation.successors[0]
    edge_data = []
    num_of_branches = (len(operations_graph)-3)
    for i in range(0, int(num_of_branches),2):
        # edge_data.append([0, i+1])
        edge_data.append(["node_0", "node_"+str(i+1)])
        # edge_data.append([i+1, i+2])
        edge_data.append(["node_"+str(i+1), "node_"+str(i+2)])
        # edge_data.append([i+2, len(operations_graph)-2])
        edge_data.append(["node_"+str(i+2), "node_"+str(len(operations_graph)-2)])

    # edge_data.append([len(operations_graph)-2, len(operations_graph)-1])
    edge_data.append(["node_"+str(len(operations_graph)-2), "node_"+str(len(operations_graph)-1)])
    print(json.dumps(output, indent=4))
    print(json.dumps(edge_data, indent=4))

def final_operation(operations_graph):
    output = []
    operation = operations_graph[0]
    while len(operation.successors) > 0:
        operation = operation.successors[0]
    print(operation)
    return operation



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, List, Union
import logging
from .utils import strip_answer_helper, strip_answer_helper_all, parse_xml


class ESCARGOTParser:
    """
    ALZKBParser provides the parsing of language model reponses specific to the
    ALZKB example.

    Inherits from the Parser class and implements its abstract methods.
    """

    def __init__(self, logger: logging.Logger = None) -> None:
        """
        Inits the response cache.
        """
        self.cache = {}
        self.logger = logger

    def parse_generate_answer(self, state: Dict, texts: List[str]) -> List[Dict]:
        """
        Parse the response from the language model for a generate prompt.

        In GOT, the generate prompt is used for planning, plan assessment, xml conversion, knowledge extraction and array function 

        :param state: The thought state used to generate the prompt.
        :type state: Dict
        :param texts: The responses to the prompt from the language model.
        :type texts: List[str]
        :return: The new thought states after parsing the respones from the language model.
        :rtype: List[Dict]
        """

        if type(texts) == str:
            texts = [texts]
        for text in texts:
            self.logger.debug(f"Got response: {text}")
            if state["method"] == "got":
                try:
                    if state["phase"] == "planning":
                        new_state = state.copy()
                        new_state["input"] = text
                        new_state["phase"] = "plan_assessment"
                        new_state["num_branches_response"] = 3
                        new_state["generate_successors"] = 1
                    elif state["phase"] == "plan_assessment":
                        new_state = state.copy()
                        #convert text to json and select from the input the top strategy
                        scores = []
                        try:
                            text = strip_answer_helper_all(text, "Approach")
                            #get all the scores
                            for i in text:
                                scores.append(int(strip_answer_helper(i,"Score")))
                            new_state["scores"] = scores
                        except Exception as e:
                            self.logger.warning(f"Could not convert text to xml: {text}. Encountered exception: {e}")
                        # new_state["phase"] = "xml_conversion"
                        new_state["phase"] = "python_conversion"
                        new_state["num_branches_response"] = 3
                        new_state["select_highest_score"] = 1
                        new_state["generate_successors"] = 1
                    elif state["phase"] == "python_conversion":
                        new_state = state.copy()
                        new_state.pop("scores")
                        new_state["input"] = text
                        self.logger.debug(f"Got Python code: {text}")
                        new_state["phase"] = "code_assessment"
                        new_state["generate_successors"] = 1
                    elif state["phase"] == "code_assessment":
                        new_state = state.copy()
                        scores = []
                        try:
                            text = strip_answer_helper_all(text, "Code")
                            for i in text:
                                scores.append(int(strip_answer_helper(i,"Score")))
                            new_state["scores"] = scores
                            # get the highest score and the approach
                            # approach = max(text, key=lambda x: int(strip_answer_helper(x,"Score")))
                            # approach = strip_answer_helper(approach, "CodeID")
                            # approach = int(approach)-1
                            # new_state["input"] = new_state["input"][approach]
                        except Exception as e:
                            self.logger.warning(f"Could not convert text to xml: {text}. Encountered exception: {e}")
                        new_state["phase"] = "xml_conversion"
                        new_state["select_highest_score"] = 1
                        new_state["generate_successors"] = 1
                        
                    elif state["phase"] == "xml_conversion":
                        new_state = state.copy()
                        new_state.pop("scores")
                        new_state["input"] = text
                        # new_state["phase"] = "xml_cleanup"
                        # new_state["generate_successors"] = 1
                        new_state["phase"] = "steps"
                        instructions, edges = parse_xml(text, self.logger)
                        new_state["instructions"] = instructions
                        new_state["edges"] = edges
                        self.logger.info("Got instructions: \n%s",instructions)
                        self.logger.info("Got edges: \n%s",edges)
                    elif state["phase"] == "xml_cleanup":
                        new_state = state.copy()
                        new_state["input"] = text
                        new_state["phase"] = "steps"
                        instructions, edges = parse_xml(text, self.logger)
                        new_state["instructions"] = instructions
                        new_state["edges"] = edges
                        self.logger.info(f"Got instructions: {instructions}")
                        self.logger.info(f"Got edges: {edges}")
                    elif state["phase"] == "steps":
                        new_state = state.copy()
                        new_state["input"] = text
                    elif state["phase"] == "output":
                        new_state = state.copy()
                        new_state["input"] = text
                except Exception as e:
                    self.logger.error(
                        f"Could not parse step answer: {text}. Encountered exception: {e}"
                    )
            
        return new_state



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, List
import re
import logging


class ESCARGOTPrompter:
    """
    ALZKBPrompter provides the generation of prompts specific to the
    ALZKB example for the language models.

    Inherits from the Prompter class and implements its abstract methods.
    """
    planning_prompt = """You are a brilliant strategic thinker with access to a knowledge base. You will receive a question that will require the knowledge base to answer. The knowledge base is built from a knowledge graph, but not the knowledge graph itself. You will break down the question into steps. If knowledge needs to be pulled from the knowledge base, you will provide what specific relationships or node information is necessary. You do not need to try to answer the question but simply plan out the steps.

The knowledge graph contains the following generic node types: {node_types}.
The knowledge graph contains the following relationships: {relationship_types}.

Only show the steps you will take and a small description for each step. If you can determine the knowledge graph relationship that can provide insight in the step, provide the relationship in it and the specific node name, not a generic node type. If a question require a specific relationship between two specific nodes, provide the specific nodes in the relationship.
No step should ask for a generic node type, such as "List all drugs" or "Find all diseases".
No step should ask for a generic relationship, such as "Determine the relationship between genes and body parts that represents over-expression", such as "GENE OVEREXPRESSES BODYPART".
Each step MUST be clear on what the knowledge request output should be. If the output is a list of genes, then the step should be clear on what genes should be returned.

Examples:
Question: List the genes which bind to the drug Cyclothiazide
Step 1: Find the genes that bind to the drug Cyclothiazide. (CHEMICALBINDSGENE)
Relationship: CHEMICALBINDSGENE
Node: Cyclothiazide
Step 2: List the genes that were found.

Question: List the genes which are commonly under-expressed in spinal cord and thyroid gland
Step 1: Identify Genes Under-Expressed in Spinal Cord. Return the genes in a list.
Relationship: BODYPARTUNDEREXPRESSESGENE
Node: Spinal Cord
Step 2: Identify Genes Under-Expressed in Thyroid Gland. Return the genes in a list.
Relationship: BODYPARTUNDEREXPRESSESGENE
Node: Thyroid Gland
Step 3: Find Common Genes
Action: Compare the lists of genes from steps 1 and 2 to identify common genes under-expressed in both the spinal cord and thyroid gland.

Question: True or False Question: Posaconazole binds to the gene CYP3A4
Step 1: Find the drugs that bind to the gene CYP3A4. List the drugs.
Relationship: GENE BINDS CHEMICAL
Node: CYP3A4
Step 2: Check if Posaconazole is in the list of drugs that bind to the gene CYP3A4.

Question: Which of the following belongs to the drug class Serotonin 3 Receptor Antagonists? 1. Dolasetron 2. Trastuzumab 3. Insulin pork 4. Hyaluronidase (ovine) 5. Streptokinase
Step 1: Find all drugs that are in the class 'Serotonin 3 Receptor Antagonists'
Relationship: DRUGCLASS
Node: Serotonin 3 Receptor Antagonists
Step 2: Check if Dolastrum, Trastuzumab, Insulin pork, Hyaluronidase (ovine), or Streptokinase are in the list of drugs in step 1.

Question: Which of the following under-expresses the genes CLTA and MARK4 ? 1. bronchus 2. sciatic nerve 3. immune system 4. internal capsule of telencephalon 5. uterus
Step 1: Find the body parts or anatomy that under-expresses the genes CLTA.
Relationship: BODYPART UNDEREXPRESSES GENE
Node: CLTA
Step 2: Find the body parts or anatomy that under-expresses the genes MARK4.
Relationship: BODYPART UNDEREXPRESSES GENE
Node: MARK4
Step 3: List the intersect of body parts

Here is your question:
{question}

Let's think step by step, and be very succinct, clear, and efficient in the number of steps by avoiding redundant knowledge extractions."""

    plan_assessment_prompt = """You are a brilliant strategic thinker with access to a biomedical knowledge graph. You will receive a query that will require the knowledge graph to answer and a few approaches that will try to resolve that query. 

If knowledge needs to be pulled from the knowledge graph, they will try to provide what specific relationships or node information is necessary.
The score should reflect on how clear and efficient each step is. It is not about how many steps are taken, but the quality of the steps. The highest score approach would be specific on what knowledge to extract, especially if it includes specific nodes. For instance, an approach with step 'Find the body parts or anatomy that over-express METTL5. (BODYPART OVEREXPRESSES GENE)' scores very high since it is specific on the node METTL5 and performs a 1 hop knowledge request.
An approach where there are steps that contain a specific node with a relationship scores higher than an approach where there any steps that contain only arbitrary node types such as when a step that is a generic node type like 'Determine the relationship between genes and body parts that represents over-expression'
Do not allow steps that simply ask for a generic node, such as "List all drugs" or "Find all diseases". These steps should be be penalized in the score.
Approaches that do not include a final step that answers the question will score lower.
If the question is a multiple choice question, the approaches that include the specific multiple choice options in at least one of the steps will score higher.
In any knowledge request, generic node types should be avoided. If a step requires a specific node, it should be used. For instance, if an approach has a step 'Find the genes that are over-expressed in the brain. (GENE OVEREXPRESSED IN BODYPART-Brain)', it will score higher than an approach that has a step 'Find all genes that are over-expressed in a body part. (GENE OVEREXPRESSED IN BODYPART)'.

The knowledge graph contains node types: {node_types}.
The knowledge graph contains relationships: {relationship_types}.

Return an XML formatted list with all the approaches in Approach tags. Each approach should be within <Approach> tags and will have an incremental <ApproachID> value within it. The score should be an integer between 1-10 within <Score> tags.
An example is as follows:
<Approaches>
  <Approach>
    <ApproachID>1</ApproachID>
    <Score>...</Score>
  </Approach>
  ...
</Approaches>

Only return the XML.

Here is your question:
{question}

Here is ApproachNumber 1:
"{approach_1}"

Here is ApproachNumber 2:
"{approach_2}"

Here is ApproachNumber 3:
"{approach_3}"
"""

    python_conversion_prompt = """You will be given a set of instructions and you must convert the instructions into Python code with the following rules:
1. Provide Python code for each step, and put the instructions for that step in comments before the code for that step. 
2. Do not define new functions. Only use default Python packages and numpy. 
3.  Do not include print statements. 
4. If the code within a step requires a request from the knowledge graph, you must use the knowledge_extract(x) function where there is only one variable in x, which is the specific knowledge you are requesting. This function will return a list. For instance, if you need to find the genes over-expressed in the brain, you would use knowledge_extract("GENE OVEREXPRESSED IN BODYPART-Brain").

Instructions:
{instructions}"""

    python_assessment_prompt = """You are a Python expert who will assess the Python code generated from the instructions. You will receive a question that requires a knowledge graph to answer and three approaches that will try to resolve that question. You will assess the quality of the Python code generated from the approaches.
You will assess the Python code based on the following criteria:
1. The code should be clear and concise.
2. The code should follow the instructions provided.
3. The code should not define any new functions.
4. The code should only use default Python packages and numpy.
5. The code should not contain placeholder variables, such as ['gene1', 'gene2', 'gene3']. Instead, the code should contain the actual variables from the instructions.
6. The code should have comments for each step that describe the instructions for that step.
7. Any code that requires the knowledge graph should use the knowledge_extract(x) function, where x is the specific knowledge you are requesting.

Return an XML formatted list with all the code snippets in Code tags. Each code snippet should be within <Code> tags and will have an incremental <CodeID> value within it. The score should be an integer between 1-10 within <Score> tags.
An example is as follows:
<Codes>
  <Code>
    <CodeID>1</CodeID>
    <Score>...</Score>
  </Code>
  ...
</Codes>

Only return the XML.

Here is the problem:
{question}
Here is your instructions:
{approach}

Here is code snippet 1:
{approach_1}

Here is code snippet 2:
{approach_2}

Here is code snippet 3:
{approach_3}
"""
    
    xml_conversion_prompt = """You will be given instructions and python code for each step. The instructions assume you have connection to a knowledge graph. The must convert it into XML with the following rules:
Format your response in XML format, where the steps will be within <Instructions> tags. Each step will be within <Step> tags and will have an incremental <StepID> value within it. The full description of the step will be put in the <Instruction> tags within the <Step>. Following the <Instruction>, you must put in the code corresponding to the step within <Code> tags.
If the code within a step requires a request from the knowledge graph, you must use the knowledge_extract(x) function where x is the specific keyword you are requesting. For instance, if you need to find the genes over-expressed in the brain, you would use knowledge_extract("GENE OVEREXPRESSED IN BODYPART-Brain"). The code will likely have placeholders for the knowledge request that you must fill in, such as ['gene1', 'gene2', 'gene3'].
Outside of the <Instructions> tag, add an edge list in <EdgeList>, where information from one step to another will be listed. Each edge will be within <Edge> tags, and the edge would be in the format StepID1-StepID2 which describes that StepID1 directs to StepID2.
Do not include any other tags other than the ones mentioned above.

Here is an example XML:
<Instructions>
    <Step>
        <StepID>1</StepID>
        <Instruction>
            Find Body Parts Over-Expressing Gene METTL5
        </Instruction>
        <Code>
            genes_overexpressed_in_nipple = knowledge_extract("BODYPART OVEREXPRESSES GENE-METTL5")
        </Code>
    </Step>
    <Step>
        <StepID>2</StepID>
        <Instruction>
            Find Body Parts Over-Expressing Gene STYXL2
        </Instruction>
        <Code>
            genes_overexpressed_in_nipple = knowledge_extract("BODYPART OVEREXPRESSES GENE-STYXL2")
        </Code>
    </Step>
    <Step>
        <StepID>3</StepID>
        <Instruction>
            List the intersect of body parts
        </Instruction>
        <Code>
            intersect = set(genes_overexpressed_in_nipple) & set(genes_overexpressed_in_brain)
        </Code>
    </Step>
</Instructions>
<EdgeList>
    <Edge>1-3</Edge>
    <Edge>2-3</Edge>
</EdgeList>

Here are the instructions you must convert:
{instructions}"""

    knowledge_request_adjustment_prompt = """You will be given a potential request extracting information from a knowledge graph, and you must convert the request into a specific format with the following rules:
1. There are specific node names, generic node types, and relationships that must be used in the query. 
2. The format of the query should be in the form of Node Name-Relationship-Node Name, where the node names can be specific nodes or generic node types.
3. If possible, use specific node names in the query, not generic node types.
4. If a specific node name is provided, surround the node name with a !. For example, if the node name is "METTL5", the query should be !METTL5!.
5. Do not surround a generic node type with a !. For example, if the node type is "Gene", the query should be Gene.
6. You must only use the the relationships and node types provided in the schema.
7. If the instruction is unclear, you must use the natural language instruction as a guide.
8. If you notice that the request does not contain a specific node name, you must use the natural language instruction as a guide.
9. Body parts should be lower case and not capitalized.
10. If the instruction has a reference to a specific step, such as "Return the genes from Step 3", you must use the step identifier in the query.
11. If there are no changes needed to the request, return the request as is.
The knowledge graph contains the following generic node types: {node_types}.
The knowledge graph contains the following relationships: {relationship_types}.

Examples:
Instruction: Identify the gene METTL5
Request: METTL5
Answer: !METTL5!

Instruction: Identify the drugs that treats Alzheimer's Disease
Request: Alzheimer's Disease-DRUG TREATS DISEASE-Drug
Answer: !Alzheimer's Disease!-DRUG TREATS DISEASE-Drug

Instruction: Find the genes that interact with the gene MCM4.
Request: GENE-GENE INTERACTS WITH GENE-MCM4
Answer: Gene-GENE INTERACTS WITH GENE-!MCM4!

Instruction: Identify Genes Under-Expressed in Brain. Return the genes in a list.
Request: GENE-BODYPART UNDER EXPRESSES GENE-Brain
Answer: Gene-BODYPART UNDER EXPRESSES GENE-!brain!

Instruction: Return the genes that are subject to decreased expression by the drug Yohimbine.
Request: Drug-CHEMICALDECREASESEXPRESSION-Gene
Answer: !Yohimbine!-CHEMICALDECREASESEXPRESSION-Gene

Instruction: Determine the gene symbol for "cytochrome c oxidase assembly factor 7"
Request: Datatype-GENE-NAME-cytochrome c oxidase assembly factor 7
Answer: Gene-!cytochrome c oxidase assembly factor 7!

Here is the instruction you must convert:
Instruction: {instruction}
Request: {statement_to_embed}

Return only the answer."""

    function_prompt="""{function}"""

    output_prompt = """Use the following knowledge to answer the question:
{knowledge}

With the above knowledge, follow this step:
{instruction}
and answer this question: {question}"""

    memgraph_prompt_1= """You are an expert memgraph Cypher translator who understands the knowledge graph request and will convert it to Cypher strictly based on the Neo4j Schema provided and following the instructions below:
1. Generate Cypher query compatible ONLY for memgraph 2.17.0
2. Do not use EXISTS, SIZE, CONTAINS ANY keywords in the cypher. Use alias when using the WITH keyword
3. Please do not use same variable names for different nodes and relationships in the query.
4. Use only Nodes and relationships mentioned in the schema
5. Always enclose the Cypher output inside 3 backticks
6. Always do a case-insensitive and fuzzy search for any properties related search. Eg: to search for a Company name use toLower(c.name) contains 'neo4j'
7. Always use aliases to refer the node in the query
8. 'Answer' is NOT a Cypher keyword. Answer should never be used in a query.
9. Please generate only one Cypher query per question. 
10. Cypher is NOT SQL. So, do not mix and match the syntaxes.
11. Every Cypher query always starts with a MATCH keyword.
12. Always use IN keyword instead of CONTAINS ANY
13. If there is a word surrounded by !, it means it is a specific node and not a node type. For instance, if the word is !Alzheimer's Disease!, it means it is a specific Disease node and not a Disease node type.
14. For the return, return only one property.
15. If a node is a Gene, please make sure you use the geneSymbol property, NOT the commonName.
16. You will receive a request for a specific node or relationship as well as a natural language instruction. Use only the specific node or relationship from the request and only if the request is unclear, use the natural language instruction as.
17. Don't worry about directionality of relationships. Assume all relationships are bidirectional and use a single dash (-) to represent relationships.
18. Return only a single property in the return statement. Do not return multiple properties.
19. When extracting information about genes, always return the geneSymbol property, not the commonName property.
20. If possible, try to keep the query to only one relationship, even if the request seems to require more than one relationship. Use the most direct relationship possible by using the instructions as a guide.
21. The natural language instruction should have what the query should return. Overwrite the return portion of the query based on the natural language instruction.

Schema:
{schema}
"""

    memgraph_prompt_2 = """
Examples:
Instruction: Identify the gene METTL5
Request: !METTL5!
Answer: MATCH (g:Gene) WHERE toLower(g.geneSymbol) = toLower("METTL5") RETURN g.geneSymbol

Instruction: Identify the drug that treats Alzheimer's Disease
Request: !Alzheimer's Disease!-DRUG TREATS DISEASE-Drug
Answer: MATCH (dr:Drug)-[:DRUGTREATSDISEASE]-(d:Disease) WHERE toLower(d.commonName) = toLower("Alzheimer's Disease") RETURN dr.commonName

Instruction: Find the genes that interact with the gene MCM4.
Request: GENE-GENE INTERACTS WITH GENE-!MCM4!
Answer: MATCH (g1:Gene {geneSymbol: "MCM4"})-[:GENEINTERACTSWITHGENE]-(g2:Gene) RETURN g2.geneSymbol

Instruction: Identify Genes Under-Expressed in Brain. Return the genes in a list.
Request: GENE-BODYPART UNDER EXPRESSES GENE-!Brain!'
Memgraph request: MATCH (bp:BodyPart)-[:BODYPARTUNDEREXPRESSESGENE]-(g:Gene) WHERE toLower(bp.commonName) = toLower("Brain") RETURN g.geneSymbol

Instruction: Find the gene(s) that are subject to decreased expression by the drug Yohimbine.
Request: Drug-CHEMICALDECREASESEXPRESSION-Gene
Memgraph request: MATCH (d:Drug {commonName: "Yohimbine"})-[:CHEMICALDECREASESEXPRESSION]-(g:Gene) RETURN g.geneSymbol

Instruction: Determine the gene symbol for "cytochrome c oxidase assembly factor 7"
Request: Datatype-GENE-NAME-!cytochrome c oxidase assembly factor 7!
Memgraph request: MATCH (g:Gene) WHERE toLower(g.commonName) = toLower("cytochrome c oxidase assembly factor 7") RETURN g.geneSymbol

"""
    memgraph_prompt_3 = """Instruction: {instruction}
Do not get confused with the above examples and return only the Cypher query for the following question: {cypher}"""

    memgraph_adjustment_prompt = """You will be given a potential request extracting information from a knowledge graph, and you must convert the request into a specific format with the following rules:
1. There are specific node names, generic node types, and relationships that must be used in the query.
2. The format of the query should be in the form of Node Name-Relationship-Node Name, where the node names can be specific nodes or generic node types.
3. If possible, use specific node names in the query, not generic node types."""

    clean_up_vector_db_prompt = """Use the following information only:
{knowledge}

Answer the question: {instruction}
Return only the answer in an array of values. Example: "[A,B,...]"""

    code_adjustment_prompt = """Use the following local variables and context to adjust the python code if you see if it gives more accurate results based on intent:
{context}

Here is the intent and the code. Do not reassign local variables and assume the local variables are in context. Adjust so that it gives accurate results:
Instruction: {instruction}
Code to adjust: {code}

Return the adjusted code only."""

    debug_code_prompt = """You will be given a Python code snippet that failed to execute. You must debug the code snippet and provide the corrected code snippet.
The instruction is as follows:
{instruction}

The code snippet is as follows:
{code}

The error message is as follows:
{error}

You must correct the code snippet and provide the corrected code snippet. Return only the code."""

    array_output_prompt = """Given the following question and instructions, you will be provided with a set of steps, associated Python code, and output for the code. You must provide the final variable that answers the question.
Question: {question}

Here are the steps, code, and output:
{steps}

Return the final variable that answers the question."""

    output_prompt = """Question:
{question}

Train of thoughts:
{steps}

Describe the thought process above, and then answer the question. Do not include any additional information or reasoning beyond what is provided.

Question:
{question}"""
    def __init__(self,vector_db = None, lm = None, memgraph_client = None, node_types = "", relationship_types = "", logger: logging.Logger = None):
        self.vector_db = vector_db
        self.lm = lm
        self.memgraph_client = memgraph_client
        self.node_types = node_types
        self.relationship_types = relationship_types
        self.logger = logger
        pass
     
    def generate_prompt(
        self,        
        question: str,
        method: str,
        input: str,
        **kwargs,
    ) -> str:
        """
        Generate a generate prompt for the language model.

        :param num_branches: The number of responses the prompt should ask the LM to generate.
        :type num_branches: int
        :param question: The question to be answered.
        :type question: str
        :param question_type: The type of the question.
        :type question_type: str
        :param method: The method used to generate the prompt.
        :type method: str
        :param input: The intermediate solution.
        :type input: str
        :param kwargs: Additional keyword arguments.
        :return: The generate prompt.
        :rtype: str
        :raise AssertionError: If method is not implemented yet.
        """
        assert question is not None, "Question should not be None."
        if method == "got":
            if (input is None or input == "") and kwargs["phase"] == "planning":
                return self.planning_prompt.format(question=question, node_types=self.node_types, relationship_types=self.relationship_types)
            elif kwargs["phase"] == "plan_assessment":
                return self.plan_assessment_prompt.format(question=question, approach_1=input[0], approach_2=input[1], approach_3=input[2], node_types=self.node_types, relationship_types=self.relationship_types)
            elif kwargs["phase"] == "python_conversion":
                return self.python_conversion_prompt.format(instructions=input)
            elif kwargs["phase"] == "code_assessment":
                return self.python_assessment_prompt.format(question=question, approach=kwargs["full_plan"], approach_1=input[0], approach_2=input[1], approach_3=input[2])
            elif kwargs["phase"] == "xml_conversion":
                return self.xml_conversion_prompt.format(instructions=input)
            elif kwargs["phase"] == "xml_cleanup":
                return self.xml_cleanup_prompt.format(xml=input, node_types=self.node_types, relationship_types=self.relationship_types)
            elif kwargs["phase"] == "output":
                steps = ""
                for step in kwargs["instructions"]:
                    steps += "Step " + step["StepID"] + ": " + step["Instruction"] + "\n"
                    steps += "Code: " + step["Code"][0] + "\n"
                    output = input[step["StepID"]]
                    if len(str(output)) > 256:
                        output = str(output)[:256] + "..."
                    steps += "Output: " + str(output) + "\n\n"
                if kwargs["answer_type"] == "array":
                    return self.array_output_prompt.format(question=question, steps=steps)
                else:
                    return self.output_prompt.format(question=question, steps=steps)
        else:
            raise AssertionError(f"Method {method} is not implemented yet.")
    def get_knowledge(self,knowledge_request,instruction):
        statement_to_embed = knowledge_request
        if statement_to_embed == "" or statement_to_embed is None:
            return []
        # statement_to_embed = self.lm.get_response_texts(
        #     self.lm.query(self.knowledge_request_adjustment_prompt.format(node_types=self.node_types,relationship_types=self.relationship_types, statement_to_embed=statement_to_embed), num_responses=1)
        # )
        # statement_to_embed = statement_to_embed[0]
        statement_to_embed_cleaned = statement_to_embed.replace("!","")
        # If it's a cypher query, then execute the query and return the results directly
        if self.memgraph_client is not None:
            knowledge_array = self.memgraph_client.execute(self.lm, self.memgraph_prompt_1.format(schema=self.memgraph_client.schema) + str(self.memgraph_prompt_2) + str(self.memgraph_prompt_3.format(instruction=str(instruction),cypher=str(statement_to_embed))),str(statement_to_embed))
            self.logger.info(f"Cypher knowledge for {statement_to_embed}: {str(knowledge_array)}")
            #backup if the memgraph client fails or doesn't provide any knowledge
            if knowledge_array == [] and self.vector_db is not None:
                statement_to_embeds = self.lm.get_response_texts(
                    self.lm.query(self.knowledge_request_adjustment_prompt.format(node_types=self.node_types,relationship_types=self.relationship_types, statement_to_embed=statement_to_embed, instruction=instruction), num_responses=3)
                )
                #get the most common element in the list
                statement_to_embed = max(set(statement_to_embeds), key = statement_to_embeds.count)
                if statement_to_embed.count("!") >= 2:
                    #if there is only once specific node and nothing else, then return the knowledge
                    embedded_question = self.lm.get_embedding(statement_to_embed_cleaned)
                    node_filters = re.findall(r'!(.*?)!', statement_to_embed)
                    knowledge_arrays = []
                    for node_filter in node_filters:
                        knowledge_array,distances = self.vector_db.get_knowledge(embedded_question, keyword_filter = node_filter)
                        if len(node_filters) > 1:
                            for node_filter in node_filters:
                                knowledge_array = [knowledge for knowledge in knowledge_array if node_filter in knowledge]
                        knowledge_arrays.append(knowledge_array)
                    knowledge_array = list(set().union(*knowledge_arrays))
                else:
                    embedded_question = self.lm.get_embedding(statement_to_embed)
                    knowledge_array,distances = self.vector_db.get_knowledge(embedded_question)
                tries = 0
                while tries < 3:
                    tries += 1
                    try:
                        knowledge_array = self.lm.get_response_texts(
                            self.lm.query(self.clean_up_vector_db_prompt.format(knowledge="\n".join(knowledge_array), instruction=instruction), num_responses=1)
                        )
                        knowledge_array = knowledge_array[0]
                        knowledge_array = eval(knowledge_array)
                        break
                    except Exception as e:
                        self.logger.error(f"Error in vector db: {e}, trying again {tries}")
                self.logger.info(f"Vector DB knowledge for {statement_to_embed}: {knowledge_array}")
            return knowledge_array
        return "knowledge"
    def generate_debug_code_prompt(self, code, instruction, e):
        prompt = self.debug_code_prompt.format(instruction=instruction, code=code, error=e)
        return prompt
    
    def adjust_code(self, code, instruction, context):
        prompt = self.code_adjustment_prompt.format(instruction=instruction, code=code, context=context)
        code = self.lm.get_response_texts(
            self.lm.query(prompt, num_responses=1)
        )[0]
        self.logger.debug(f"Adjusting code prompt: {prompt}")
        self.logger.debug(f"Adjusted code: {code}")
        return code




import weaviate
# from embed import get_embedding
from VectorEmbeddingSearchingNick_v1.embed import get_embedding
import os
from openai import AzureOpenAI
import numpy as np
import tiktoken

weaviate_client = None


def get_token_length(text, model="gpt-3.5-turbo"):
    tokenizer = tiktoken.encoding_for_model(model)
    tokens = tokenizer.encode(text)
    return len(tokens)

def calculate_message_length( message_text, model="gpt-3.5-turbo"):
        
        
        total_length = sum(get_token_length(msg['content'], model) for msg in message_text)
        return total_length

# def get_client(url="http://54.203.49.88"):
def get_client(url="http://100.21.60.48", weaviate_api_key = None):   
    auth_config = weaviate.AuthApiKey(api_key=weaviate_api_key)
    global weaviate_client
    if weaviate_client is None:
        weaviate_client = weaviate.Client(
            url=url,
            auth_client_secret=auth_config
        )
    return weaviate_client


def query_bm25(class_name="AlzKB", properties=["knowledge"], query_string="test", additional="score", limit=3):
    return ((get_client().query
        .get(class_name, properties))
        .with_bm25(
            query=query_string,
            properties=properties  # this does not need to be the same as columns
        )
        .with_additional(additional)
        .with_limit(limit)
        .do()
    )


def query_near_text(class_name="AlzKB", properties=["knowledge"], near_text=["gene"], additional="score", limit=3):
    return ((get_client().query
        .get(class_name, properties))
        .with_near_text({
            "concepts": near_text
        })
        .with_limit(limit)
        .with_additional(additional)
        .do()
    )


def query_my_near_text(class_name="AlzKB", properties=["knowledge"], prompt="genes", additional="score", limit=3, config_RAG_TEST = None):
    vector = {
        "vector": get_embedding(prompt, config_RAG_TEST)
    }
    # print(vector["vector"])
    return query_near_vector(class_name, properties, near_vector=vector, additional=additional, limit=limit, config_RAG_TEST=config_RAG_TEST)

def query_near_vector(class_name="AlzKB", properties=["knowledge"], near_vector={}, additional="score", limit=3, config_RAG_TEST = None):
    WEAVIATE_APIKEY = config_RAG_TEST["weaviate"]["api_key"]
    # weaviate url
    weaviate_url = config_RAG_TEST["weaviate"]["url"]
    return ((get_client(url=weaviate_url ,weaviate_api_key = WEAVIATE_APIKEY).query
        .get(class_name, properties))
        .with_near_vector(near_vector)
        .with_limit(limit)
        .with_additional(additional)
        .do()
    )
    
def object_count(class_name="AlzKB"):
    return ((get_client().query
        .aggregate(class_name)
        .with_meta_count()
        .do()))
    
def get_answer(question, knowledge_array, config_RAG_TEST = None):
    

    # Azure 
    AZURE_OPENAI_KEY = config_RAG_TEST["azuregpt35-16k"]["api_key"]

    
    model = config_RAG_TEST["azuregpt35-16k"]["model_id"]
    print("model: ", model)

    # temperature_azure_gpt = 0
    # temperature_azure_gpt = 0.7
    temperature = config_RAG_TEST["azuregpt35-16k"]["temperature"]

    # max_tokens_azure_gpt = 1000
    max_tokens = config_RAG_TEST["azuregpt35-16k"]["max_tokens"]

    # api_version
    api_version = config_RAG_TEST["azuregpt35-16k"]["api_version"]


    # AZURE_OPENAI_ENDPOINT=https://caire-azure-openai.openai.azure.com/openai/deployments/gpt-35-turbo-16k/chat/completions?api-version=2023-07-01-preview

    AZURE_OPENAI_ENDPOINT = config_RAG_TEST["azuregpt35-16k"]["api_base"]+"api-version="+api_version


    # if model is gpt-35-turbo-16k then max
    if model == "gpt-35-turbo-16k":
        max_context_length = 16384
    

    in_context = "\n".join(knowledge_array)

    # print length of in_context
    print("length of in_context: ", len(in_context))
    client = AzureOpenAI(
        api_key=AZURE_OPENAI_KEY,
        api_version=api_version,
        azure_endpoint = AZURE_OPENAI_ENDPOINT
        )
        

    
    message_text = [{"role":"system","content":"""You are an Alzheimer's data specialist AI assistant dedicated to providing information and support related to Alzheimer's disease.
        Your primary goal is to assist users by offering factual and relevant information based on your access to a comprehensive knowledge graph associated with Alzheimer's. 
        Your responses are focused on addressing queries related to Alzheimer's, and you do not provide information unrelated to the topic. 
        You will also only answer based on the knowledge within the knowledge graph. 
        You will notice there will be gene symbols in the knowledge, and there are subtle differences between the gene names.
        You will need to be careful that the names are exact with you use them in context. There may be single differences in numbers and letters.
        For example, the gene "APOE" is not the same as gene "APOE1". Another example is the gene "IQCK" is not the same as gene "IQCG".
        You will need to be careful of specific biological terms. For example, the term "amino" is different from the term "amine".
        If you are providing a list, be sure not to list duplicates. 
        Your demeanor is empathetic and concise as you aim to help users understand and navigate Alzheimer's-related concerns."""},
        {"role":"user","content":"Here is the knowledge from the knowledge graph: "+in_context+"\nThe question is: "+question}]
    

    # initial message length check
    message_length = calculate_message_length(message_text)
    print("Initial token length of message_text: ", message_length)

    # Reduce in_context size until it fits within max_context_length
    while message_length + max_tokens > max_context_length:
        # Reduce in_context size
        knowledge_array = knowledge_array[:-1]
        in_context = "\n".join(knowledge_array)


        message_text = [{"role":"system","content":"""You are an Alzheimer's data specialist AI assistant dedicated to providing information and support related to Alzheimer's disease.
        Your primary goal is to assist users by offering factual and relevant information based on your access to a comprehensive knowledge graph associated with Alzheimer's. 
        Your responses are focused on addressing queries related to Alzheimer's, and you do not provide information unrelated to the topic. 
        You will also only answer based on the knowledge within the knowledge graph. 
        You will notice there will be gene symbols in the knowledge, and there are subtle differences between the gene names.
        You will need to be careful that the names are exact with you use them in context. There may be single differences in numbers and letters.
        For example, the gene "APOE" is not the same as gene "APOE1". Another example is the gene "IQCK" is not the same as gene "IQCG".
        You will need to be careful of specific biological terms. For example, the term "amino" is different from the term "amine".
        If you are providing a list, be sure not to list duplicates. 
        Your demeanor is empathetic and concise as you aim to help users understand and navigate Alzheimer's-related concerns."""},
        {"role":"user","content":"Here is the knowledge from the knowledge graph: "+in_context+"\nThe question is: "+question}]
        

        message_length = calculate_message_length(message_text)
        print("Reduced token length of message_text: ", message_length)
    
    # text that may help:
    # When you reply, please provide a followup response that includes the exact knowledge from the knowledge graph that you used to generate your response, and please do not include the knowledge that are not used.
        
    response = client.chat.completions.create(
        model=model, # model = "deployment_name".
        max_tokens=max_tokens,
        messages=message_text,
        temperature=temperature
    )
    # response = client.completions.create(model=deployment_name, prompt=message_text, max_tokens=100)
    # print(response)
    # print(response.choices[0].message.content)
    return response.choices[0].message.content


def get_answer_without_ICL_Azure_gpt_35_turbo_16k(question, config_RAG_TEST = None):
    temperature = config_RAG_TEST["azuregpt35-16k"]["temperature"]
    max_tokens = config_RAG_TEST["azuregpt35-16k"]["max_tokens"]

    # in_context = "\n".join(knowledge_array)
    client = AzureOpenAI(
        api_key=config_RAG_TEST["azuregpt35-16k"]["api_key"],
        api_version=config_RAG_TEST["azuregpt35-16k"]["api_version"],
        azure_endpoint = config_RAG_TEST["azuregpt35-16k"]["api_base"]+"api-version="+config_RAG_TEST["azuregpt35-16k"]["api_version"]
        )
        
    deployment_name=config_RAG_TEST["azuregpt35-16k"]["model_id"] #This will correspond to the custom name you chose for your deployment when you deployed a model.
    
    
    message_text = [{"role":"system","content":"""You are an Alzheimer's data specialist AI assistant dedicated to providing information and support related to Alzheimer's disease.
        Your primary goal is to assist users by offering factual and relevant information based on your access to a comprehensive knowledge graph associated with Alzheimer's. 
        Your responses are focused on addressing queries related to Alzheimer's, and you do not provide information unrelated to the topic. 
        You will also only answer based on the knowledge within the knowledge graph. 
        You will notice there will be gene symbols in the knowledge, and there are subtle differences between the gene names.
        You will need to be careful that the names are exact with you use them in context. There may be single differences in numbers and letters.
        For example, the gene "APOE" is not the same as gene "APOE1". Another example is the gene "IQCK" is not the same as gene "IQCG".
        You will need to be careful of specific biological terms. For example, the term "amino" is different from the term "amine".
        If you are providing a list, be sure not to list duplicates. 
        Your demeanor is empathetic and concise as you aim to help users understand and navigate Alzheimer's-related concerns."""},
                    # {"role":"user","content":"Here is the knowledge from the knowledge graph: "+in_context+"\nThe question is: "+question}]
                    {"role":"user","content":"The question is: "+question}]
    
    # text that may help:
    # When you reply, please provide a followup response that includes the exact knowledge from the knowledge graph that you used to generate your response, and please do not include the knowledge that are not used.
        
    response = client.chat.completions.create(
        model=deployment_name, # model = "deployment_name".
        max_tokens=max_tokens,
        messages=message_text,
        temperature=temperature



    )
    # response = client.completions.create(model=deployment_name, prompt=message_text, max_tokens=100)
    # print(response)
    # print(response.choices[0].message.content)
    return response.choices[0].message.content

def get_answer_without_ICL_Azure_gpt_4(question,temperature = 0, max_tokens=1000):
    # in_context = "\n".join(knowledge_array)

    # AZURE_OPENAI_KEY_GPT_4=035a7cc7324d4e01a3f71d25dfb2165e
    # AZURE_OPENAI_ENDPOINT_GPT_4=https://caire-gpt4.openai.azure.com/
    # api-version_GP4=2024-02-15-preview

    client = AzureOpenAI(

        # api_key=os.environ["AZURE_OPENAI_KEY_GPT_4"],
        # api_version=os.environ["api-version_GP4"],
        # azure_endpoint = os.environ["AZURE_OPENAI_ENDPOINT_GPT_4"]
        api_key="035a7cc7324d4e01a3f71d25dfb2165e",
        api_version="2024-02-15-preview",
        azure_endpoint = "https://caire-gpt4.openai.azure.com/"
        )
        
    deployment_name='gpt-4' #This will correspond to the custom name you chose for your deployment when you deployed a model.
    
    
    message_text = [{"role":"system","content":"""You are an Alzheimer's data specialist AI assistant dedicated to providing information and support related to Alzheimer's disease.
        Your primary goal is to assist users by offering factual and relevant information based on your access to a comprehensive knowledge graph associated with Alzheimer's. 
        Your responses are focused on addressing queries related to Alzheimer's, and you do not provide information unrelated to the topic. 
        You will also only answer based on the knowledge within the knowledge graph. 
        You will notice there will be gene symbols in the knowledge, and there are subtle differences between the gene names.
        You will need to be careful that the names are exact with you use them in context. There may be single differences in numbers and letters.
        For example, the gene "APOE" is not the same as gene "APOE1". Another example is the gene "IQCK" is not the same as gene "IQCG".
        You will need to be careful of specific biological terms. For example, the term "amino" is different from the term "amine".
        If you are providing a list, be sure not to list duplicates. 
        Your demeanor is empathetic and concise as you aim to help users understand and navigate Alzheimer's-related concerns."""},
                    # {"role":"user","content":"Here is the knowledge from the knowledge graph: "+in_context+"\nThe question is: "+question}]
                    {"role":"user","content":"The question is: "+question}]
    
    # text that may help:
    # When you reply, please provide a followup response that includes the exact knowledge from the knowledge graph that you used to generate your response, and please do not include the knowledge that are not used.
        
    response = client.chat.completions.create(
        model=deployment_name, # model = "deployment_name".
        max_tokens=max_tokens,
        messages=message_text,
        temperature=temperature
    )
    # response = client.completions.create(model=deployment_name, prompt=message_text, max_tokens=100)
    # print(response)
    # print(response.choices[0].message.content)
    return response.choices[0].message.content


# bm25 (keyword) search
# result = query_bm25(query_string="list all genes connected to alzheimers", limit=20)
# result = query_bm25(query_string="TPSAB1", additional="vector")
# result = query_bm25(query_string="TPSAB1", additional="id")
# result = query_bm25(query_string="TPSAB1", additional="distance") # returns: None

# # Similarity / Vector search
# near_vector = {
#     # what genes code for acetyl-CoA carboxylase?
# }
# result = query_near_vector(near_vector=near_vector, additional="distance", limit=10)

# My nearText search
# print(os.environ)

def condense_knowledge(knowledge_array):
    in_context = "\n".join(knowledge_array)
    client = AzureOpenAI(
        api_key=os.environ["OPENAI_KEY"],
        api_version="2023-07-01-preview",
        azure_endpoint = "https://caire-azure-openai.openai.azure.com/openai/deployments/gpt-35-turbo-16k/chat/completions?api-version=2023-07-01-preview"
        )
        
    deployment_name='gpt-35-turbo-16k' #This will correspond to the custom name you chose for your deployment when you deployed a model. 
    
    message_text = [{"role":"system","content":"""You are a data specialist AI assistant dedicated to providing a more concise version of the knowledge graph associated with Alzheimer's disease.
                     You will receive a fragmented list of knowledge from the knowledge graph, and you will need to condense the knowledge into a more concise version.
                     Be sure to include all the knowledge from the knowledge graph, and do not include knowledge that are not in the knowledge graph.
                     You will notice there will be gene symbols in the knowledge, and there are subtle differences between the gene names.
                     You will need to be careful that the names are exact with you use them in context. There may be single differences in numbers and letters.
                     For example, the gene "APOE" is not the same as gene "APOE1". Another example is the gene "IQCK" is not the same as gene "IQCG".
                     You will need to be careful of specific biological terms. For example, the term "amino" is different from the term "amine".
                     If you are providing a list, be sure not to list duplicates."""},
                    {"role":"user","content":"Here is the knowledge from the knowledge graph: "+in_context}]
    
    # text that may help:
    # When you reply, please provide a followup response that includes the exact knowledge from the knowledge graph that you used to generate your response, and please do not include the knowledge that are not used.
        
    response = client.chat.completions.create(
        model=deployment_name, # model = "deployment_name".
        max_tokens=1000,
        messages=message_text
    )
    return response.choices[0].message.content

def compare_answers(answer, ground_truth):
    client = AzureOpenAI(
        api_key=os.environ["OPENAI_KEY"],
        api_version="2023-07-01-preview",
        azure_endpoint = "https://caire-azure-openai.openai.azure.com/openai/deployments/gpt-35-turbo-16k/chat/completions?api-version=2023-07-01-preview"
        )
        
    deployment_name='gpt-35-turbo-16k' #This will correspond to the custom name you chose for your deployment when you deployed a model. 
    
    message_text = [{"role":"system","content":"""You are a data specialist AI that is comparing an answer to a ground truth.
                     You will receive an answer and a ground truth, and you will need to compare the answer to the ground truth.
                     In the end, you will need to provide a score between 0 and 1, where 0 is the worst score and 1 is the best score.
                     You will need to be careful that the names are exact with you use them in context. There may be single differences in numbers and letters.
                     For example, the gene "APOE" is not the same as gene "APOE1". Another example is the gene "IQCK" is not the same as gene "IQCG".
                     You will need to be careful of specific biological terms. For example, the term "amino" is different from the term "amine".
                     If you are comparing lists, be very sure that the lists are exactly the same and understand that it may not be in the same order."""},
                    {"role":"user","content":"Here is the answer: "+answer+"\nHere is the ground truth: "+ground_truth}]
    
    # text that may help:
    # When you reply, please provide a followup response that includes the exact knowledge from the knowledge graph that you used to generate your response, and please do not include the knowledge that are not used.
        
    response = client.chat.completions.create(
        model=deployment_name, # model = "deployment_name".
        max_tokens=1000,
        messages=message_text
    )
    return response.choices[0].message.content

def get_score_from_comparison(comparison):
    client = AzureOpenAI(
        api_key=os.environ["OPENAI_KEY"],
        api_version="2023-07-01-preview",
        azure_endpoint = "https://caire-azure-openai.openai.azure.com/openai/deployments/gpt-35-turbo-16k/chat/completions?api-version=2023-07-01-preview"
        )
        
    deployment_name='gpt-35-turbo-16k' #This will correspond to the custom name you chose for your deployment when you deployed a model. 
    
    message_text = [{"role":"system","content":"""You are a data specialist AI that is given a description of comparing an answer to the truth and possibly a numeric score.
                     You will extract the score from the description and output only the score.
                     Your output should not include any other text. For example, if the description is "The score is 0.5", you should output "0.5".
                     If there is no score, you will output nothing.
                     If there is no score, but determine that the description says it is a good answer, you will output "1"."""},
                    {"role":"user","content":"Here is the description: "+comparison}]
    
    # text that may help:
    # When you reply, please provide a followup response that includes the exact knowledge from the knowledge graph that you used to generate your response, and please do not include the knowledge that are not used.
        
    response = client.chat.completions.create(
        model=deployment_name, # model = "deployment_name".
        max_tokens=1000,
        messages=message_text
    )
    return response.choices[0].message.content

def get_knowledge(question, config_RAG_TEST = None):

    # weaviate
    max_tokens = config_RAG_TEST["weaviate"]["max_tokens_weaviate"]
    quantile = config_RAG_TEST["weaviate"]["quantile_weaviate"]
    max_distance = config_RAG_TEST["weaviate"]["max_distance_weaviate"]
    

    

    # print question, max_tokens, quantile, max_distance, show_distances using print function
    # print("question: ", question)
    # print("max_tokens: ", max_tokens)
    # print("quantile: ", quantile)
    # print("max_distance: ", max_distance)
    # print("show_distances: ", show_distances)

    knowledge_array = query_my_near_text(prompt=question, additional=["distance"], limit=2000,config_RAG_TEST=config_RAG_TEST)

    knowledge_array = knowledge_array["data"]["Get"]["AlzKB"]
    
    all_distances = [x['_additional']["distance"] for x in knowledge_array]
    quant_dist = np.quantile(all_distances, quantile)
    in_context = []
    distances = []
    cur_tokens = 0
    for knowledge in knowledge_array:
        # print(knowledge['_additional']["distance"])
        # print(knowledge["knowledge"])
        if knowledge['_additional']["distance"] > quant_dist or knowledge['_additional']["distance"] > max_distance:
            break
        
        cur_tokens += len(knowledge["knowledge"].split(" "))
        if cur_tokens > max_tokens:
            break
        in_context.append(knowledge["knowledge"])
        distances.append(knowledge['_additional']["distance"])
        
    return in_context, distances

# import keys



"""

if __name__ == '__main__':
    # bm25 (keyword) search
    # result = query_bm25(query_string="list all genes connected to alzheimers", limit=20)
    # result = query_bm25(query_string="TPSAB1", additional="vector")
    # result = query_bm25(query_string="TPSAB1", additional="id")
    # result = query_bm25(query_string="TPSAB1", additional="distance") # returns: None

    # # Similarity / Vector search
    # near_vector = {
    #     # what genes code for acetyl-CoA carboxylase?
    # }
    # result = query_near_vector(near_vector=near_vector, additional="distance", limit=10)

    # My nearText search
    # print(os.environ)


    question = "what drugs not associated with alzheimer's disease may be effective in treating alzheimer's disease and why?"

    # quantile = 0.25 # 500
    # quantile = 0.15 # 300
    # quantile = 0.1 # 200
    # quantile = 0.05 # 100
    # quantile = 0.01 # 20
    quantile = 0.005 # 10

    knowledge_array,distances = get_knowledge(question, max_tokens=10000, quantile = quantile, max_distance=0.25, show_distances = True)
    print("knowledge array length: ", len(knowledge_array))
    # knowledge = condense_knowledge(knowledge_array)
    answer = get_answer(question,knowledge_array,temperature = 0)
    print(answer)

"""



from __future__ import annotations
from typing import Iterator, Dict, Optional
import itertools


class Thought:
    """
    Represents an LLM thought with its state, constructed by the parser, and various flags.
    """

    _ids: Iterator[int] = itertools.count(0)

    def __init__(self, state: Optional[Dict] = None) -> None:
        """
        Initializes a new Thought instance with a state and various default flags.

        :param state: The state of the thought. Defaults to None.
        :type state: Optional[Dict]
        """
        self.id: int = next(Thought._ids)
        self.state: Dict = state


# Copyright (c) 2023 ETH Zurich.
#                    All rights reserved.
#
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
#
# main author: Nils Blach

import json
import logging
from typing import List, Optional, Dict, Any
from escargot.language_models import AbstractLanguageModel
from escargot.operations import GraphOfOperations, Thought
from escargot.prompter import ESCARGOTPrompter
from escargot.parser import ESCARGOTParser
from escargot.coder import Coder
import copy

class Controller:
    """
    Controller class to manage the execution flow of the Graph of Operations,
    generating the Graph Reasoning State.
    This involves language models, graph operations, prompting, and parsing.
    """

    def __init__(
        self,
        lm: AbstractLanguageModel,
        graph: GraphOfOperations,
        prompter: ESCARGOTPrompter,
        parser: ESCARGOTParser,
        logger: logging.Logger,
        problem_parameters: Dict[str, Any],
    ) -> None:
        """
        Initialize the Controller instance with the language model,
        operations graph, prompter, parser, and problem parameters.
        """
        self.logger = logger
        self.lm = lm
        self.graph = graph
        self.prompter = prompter
        self.parser = parser
        self.original_problem_parameters = problem_parameters
        self.problem_parameters = problem_parameters
        self.run_executed = False
        self.got_steps = {}
        self.final_thought = None
        self.execution_queue = []
        self.max_run_tries = 3
        self.max_operation_tries = 2
        self.coder = Coder()

    def initialize_execution_queue(self) -> None:
        """
        Initialize the execution queue with the root operations of the graph.
        """
        self.execution_queue = [
            operation for operation in self.graph.operations if operation.can_be_executed()
        ]

    def execute_step(self) -> Optional[Thought]:
        """
        Execute one step from the execution queue.

        :return: The thought generated by the executed operation, or None if no operation is left to execute.
        """
        if not self.execution_queue:
            self.initialize_execution_queue()

        if not self.execution_queue:
            self.logger.debug("No more operations to execute.")
            return None

        current_operation = self.execution_queue.pop(0)
        current_operation_backup = copy.copy(current_operation)
        
        tries = 0
        while tries < self.max_operation_tries:
            try:
                current_operation.execute(
                    self.lm, self.prompter, self.parser, self.got_steps, self.logger, self.coder, **self.problem_parameters
                )
                break
            except Exception as e:
                self.logger.error("Error executing operation %s: %s", current_operation.operation_type, e)
                current_operation = copy.copy(current_operation_backup)
                tries += 1

        del current_operation_backup

        if tries == self.max_operation_tries:
            self.logger.error("Max tries reached on executing operation %s", current_operation.operation_type)
            return None
        
        for operation in current_operation.successors:
            if operation.can_be_executed():
                self.execution_queue.append(operation)
        
        self.final_thought = current_operation.get_thoughts()[0]
        
        return self.final_thought

    def run(self) -> None:
        """
        Run the controller and execute the operations from the Graph of Operations based on their readiness.
        Ensures the program is in a valid state before execution.
        """
        assert self.graph.roots is not None, "The operations graph has no root"
        
        while not self.run_executed and self.max_run_tries > 0:
            self.max_run_tries -= 1
            self.execution_queue = []
            self.got_steps = {}
            self.problem_parameters = copy.copy(self.original_problem_parameters)
            self.initialize_execution_queue()
            
            while self.execution_queue:
                self.execute_step()

            if self.final_thought.state["phase"] == "output":
                self.logger.info("All operations executed")
                self.run_executed = True

        if self.max_run_tries == 0:
            self.logger.error("Max tries reached on executing controller")


    def get_final_thoughts(self) -> List[List[Thought]]:
        """
        Retrieve the final thoughts after all operations have been executed.

        :return: List of thoughts for each operation in the graph's leaves.
        """
        assert self.run_executed, "The run method has not been executed"
        return [operation.get_thoughts() for operation in self.graph.leaves]

    def serialize_operation(self, operation) -> Dict[str, Any]:
        """
        Serialize an operation to a dictionary.

        :param operation: The operation to serialize.
        :return: The serialized operation.
        """
        operation_serialized = {
            "operation": operation.operation_type.name,
            "thoughts": [thought.state for thought in operation.get_thoughts()],
        }
        return operation_serialized

    def output_graph(self, path: str) -> None:
        """
        Serialize the state and results of the operations graph to a JSON file.

        :param path: The path to the output file.
        """
        output = [self.serialize_operation(op) for op in self.graph.operations]

        output.append(
            {
                "prompt_tokens": self.lm.prompt_tokens,
                "completion_tokens": self.lm.completion_tokens,
                "cost": self.lm.cost,
            }
        )

        with open(path, "w") as file:
            json.dump(output, file, indent=2)


import backoff
import os
import random
import time
from typing import List, Dict, Union
import openai
from openai import AzureOpenAI, OpenAIError
from openai.types.chat.chat_completion import ChatCompletion
import logging

from .abstract_language_model import AbstractLanguageModel


class AzureGPT(AbstractLanguageModel):
    """
    The ChatGPT class handles interactions with the OpenAI models using the provided configuration.

    Inherits from the AbstractLanguageModel and implements its abstract methods.
    """

    def __init__(
        self, config_path: str = "", model_name: str = "chatgpt", cache: bool = False, logger: logging.Logger = None
    ) -> None:
        """
        Initialize the ChatGPT instance with configuration, model details, and caching options.

        :param config_path: Path to the configuration file. Defaults to "".
        :type config_path: str
        :param model_name: Name of the model, default is 'chatgpt'. Used to select the correct configuration.
        :type model_name: str
        :param cache: Flag to determine whether to cache responses. Defaults to False.
        :type cache: bool
        """
        super().__init__(config_path, model_name, cache, logger)
        self.config: Dict = self.config[model_name]
        # The model_id is the id of the model that is used for chatgpt, i.e. gpt-4, gpt-3.5-turbo, etc.
        self.model_id: str = self.config["model_id"]
        # The embedding_id is the id of the embedding model that is used
        self.embedding_id: str = self.config["embedding_id"]
        # The prompt_token_cost and response_token_cost are the costs for 1000 prompt tokens and 1000 response tokens respectively.
        self.prompt_token_cost: float = self.config["prompt_token_cost"]
        self.response_token_cost: float = self.config["response_token_cost"]
        # The temperature of a model is defined as the randomness of the model's output.
        self.temperature: float = self.config["temperature"]
        # The maximum number of tokens to generate in the chat completion.
        self.max_tokens: int = self.config["max_tokens"]
        # The stop sequence is a sequence of tokens that the model will stop generating at (it will not generate the stop sequence).
        self.stop: Union[str, List[str]] = self.config["stop"]
        self.api_base: str = self.config["api_base"]
        if self.api_base == "":
            self.logger.warning("OPENAI_API_BASE is not set")
        self.api_version: str = self.config["api_version"]
        if self.api_version == "":
            self.logger.warning("OPENAI_API_VERSION is not set")
        self.api_key: str = os.getenv("OPENAI_API_KEY", self.config["api_key"])
        if self.api_key == "":
            raise ValueError("OPENAI_API_KEY is not set")
        self.response_cache: Dict[str, List[ChatCompletion]] = {}
        # Initialize the OpenAI Client
        self.client = AzureOpenAI(api_key=self.api_key,azure_endpoint=self.api_base,api_version=self.api_version)

    def query(
        self, query: str, num_responses: int = 1
    ) -> Union[List[ChatCompletion], ChatCompletion]:
        """
        Query the OpenAI model for responses.

        :param query: The query to be posed to the language model.
        :type query: str
        :param num_responses: Number of desired responses, default is 1.
        :type num_responses: int
        :return: Response(s) from the OpenAI model.
        :rtype: Dict
        """
        if self.cache and query in self.response_cache:
            return self.response_cache[query]

        response = []
        response = self.chat([{"role": "system", "content": query}], num_responses)
        
        if self.cache:
            self.response_cache[query] = response
        return response

    @backoff.on_exception(backoff.expo, OpenAIError, max_time=10, max_tries=6)
    def chat(self, messages: List[Dict], num_responses: int = 1) -> ChatCompletion:
        """
        Send chat messages to the OpenAI model and retrieves the model's response.
        Implements backoff on OpenAI error.

        :param messages: A list of message dictionaries for the chat.
        :type messages: List[Dict]
        :param num_responses: Number of desired responses, default is 1.
        :type num_responses: int
        :return: The OpenAI model's response.
        :rtype: ChatCompletion
        """

        #TODO: Optimize temperature, max_tokens, and stop
        response = self.client.chat.completions.create(
            model=self.model_id,
            messages=messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            n=num_responses,
            # stop=self.stop,
        )

        self.prompt_tokens += response.usage.prompt_tokens
        self.completion_tokens += response.usage.completion_tokens
        prompt_tokens_k = float(self.prompt_tokens) / 1000.0
        completion_tokens_k = float(self.completion_tokens) / 1000.0
        self.cost = (
            self.prompt_token_cost * prompt_tokens_k
            + self.response_token_cost * completion_tokens_k
        )
        return response

    def get_response_texts(
        self, query_response: Union[List[ChatCompletion], ChatCompletion]
    ) -> List[str]:
        """
        Extract the response texts from the query response.

        :param query_response: The response dictionary (or list of dictionaries) from the OpenAI model.
        :type query_response: Union[List[ChatCompletion], ChatCompletion]
        :return: List of response strings.
        :rtype: List[str]
        """
        if not isinstance(query_response, List):
            query_response = [query_response]
        return [
            choice.message.content
            for response in query_response
            for choice in response.choices
        ]
    
    def get_embedding(self, text_to_embed):

        response = self.client.embeddings.create(
            model=self.embedding_id,
            input=text_to_embed
        )

        return response.data[0].embedding



import re
import logging
import json
import xml.etree.ElementTree as ET
import logging

def strip_answer_helper(text: str, tag: str = "") -> str:
    """
    Helper function to remove tags from a text.

    :param text: The input text.
    :type text: str
    :param tag: The tag to be stripped. Defaults to "".
    :type tag: str
    :return: The stripped text.
    :rtype: str
    """

    text = text.strip()
    if "Output:" in text:
        text = text[text.index("Output:") + len("Output:") :].strip()
    if tag != "":
        start = text.rfind(f"<{tag}>")
        end = text.rfind(f"</{tag}>")
        if start != -1 and end != -1:
            text = text[start + len(f"<{tag}>") : end].strip()
        elif start != -1:
            logging.warning(
                f"Only found the start tag <{tag}> in answer: {text}. Returning everything after the tag."
            )
            text = text[start + len(f"<{tag}>") :].strip()
        elif end != -1:
            logging.warning(
                f"Only found the end tag </{tag}> in answer: {text}. Returning everything before the tag."
            )
            text = text[:end].strip()
        else:
            logging.warning(
                f"Could not find any tag {tag} in answer: {text}. Returning the full answer."
            )
    return text

#strip answer helper but returns all instances of the tag
def strip_answer_helper_all(text: str, tag: str = "") -> str:
    """
    Helper function to remove tags from a text.

    :param text: The input text.
    :type text: str
    :param tag: The tag to be stripped. Defaults to "".
    :type tag: str
    :return: The stripped text.
    :rtype: str
    """

    text = text.strip()
    #get all instances of the tag
    start = [m.start() for m in re.finditer(f"<{tag}>", text)]
    end = [m.start() for m in re.finditer(f"</{tag}>", text)]
    # print(start)
    # print(end)
    return [text[text.index(f"<{tag}>", start[i]) + len(f"<{tag}>") : end[i]].strip() for i in range(len(start))]

def parse_xml(xml_data, logger):
    # Parse the XML string
    #find <?xml version="1.0" encoding="UTF-8"?> and remove it
    xml_data = re.sub(r"<\?xml version=\"1.0\" encoding=\"UTF-8\"\?>", "", xml_data)
    #find ```xml and remove it
    xml_data = re.sub(r"```xml", "", xml_data)
    #find ``` and remove it
    xml_data = re.sub(r"```", "", xml_data)
    #find <Root> and remove it
    xml_data = re.sub(r"<Root>", "", xml_data)
    #find </Root> and remove it
    xml_data = re.sub(r"</Root>", "", xml_data)
    xml_data = xml_data.replace("&", "&amp;")
    try:
        xml_data = '<Root>' + xml_data + '</Root>'
        root = ET.fromstring(xml_data)
    except Exception as e:
        logger.error(f"Could not parse XML data: {xml_data}. Encountered exception: {e}")
        return None, None

    def get_step(step):
        step_id = step.find('StepID').text
        instruction = step.find('Instruction')

        if step is None:
            return None  # Handle cases where there's no instruction element
        
        # Initialize empty lists to store information
        codes = []
        for info in step.findall('Code'):
            code_text = info.text.strip()
            if code_text:
                #unescape the code
                code_text = code_text.replace("&amp;", "&")
            codes.append(code_text)

        # Return a list with relevant information (adjust as needed)
        return {
            "StepID": step_id,
            "Instruction": instruction.text.strip() if instruction.text else "",
            "Code": codes
        }

    # print('xml_data:',xml_data.split("\n"))
    # Extract and print details for each step
    instructions = root.find('Instructions').findall('Step')
    steps = []
    for step in instructions:
        parsed_step = get_step(step)
        # print(parsed_step)
        steps.append(parsed_step)

    # Extract and print edges
    #check if EdgeList exists
    if root.find('EdgeList') is None:
        return steps, []
    edges = root.find('EdgeList').findall('Edge')
    #remove \n from the text and whitespace
    edges = [edge.text.replace("\n","").strip() for edge in edges]
    # for edge in edges:
    #     print(f'Edge: {edge.text}')
    # print("edges:", edges)
    return steps, edges

def output_controller(operations_graph):

    output = []
    # for operation in operations_log['MCQ_1hop.json']['Which of the following binds to the drug Leucovorin? 1. CAD 2. PDS5B 3. SEL1L 4. ABCC2 5. RMI1']:
    # print(len(operations_graph))
    index = 0
    operation = operations_graph[0]
    while len(operation.successors) > 0:
    # for operation in operations_graph:
        
        operation_serialized = {
            "id": "node_"+str(index),
            "operation": operation.operation_type.name,
            "thoughts": [thought.state for thought in operation.get_thoughts()],
        }
        print(operation_serialized["thoughts"][0]["prompt"])
        
        output.append(operation_serialized)
        index = index + 1
        operation = operation.successors[0]
    edge_data = []
    num_of_branches = (len(operations_graph)-3)
    for i in range(0, int(num_of_branches),2):
        # edge_data.append([0, i+1])
        edge_data.append(["node_0", "node_"+str(i+1)])
        # edge_data.append([i+1, i+2])
        edge_data.append(["node_"+str(i+1), "node_"+str(i+2)])
        # edge_data.append([i+2, len(operations_graph)-2])
        edge_data.append(["node_"+str(i+2), "node_"+str(len(operations_graph)-2)])

    # edge_data.append([len(operations_graph)-2, len(operations_graph)-1])
    edge_data.append(["node_"+str(len(operations_graph)-2), "node_"+str(len(operations_graph)-1)])
    print(json.dumps(output, indent=4))
    print(json.dumps(edge_data, indent=4))

def final_operation(operations_graph):
    output = []
    operation = operations_graph[0]
    while len(operation.successors) > 0:
        operation = operation.successors[0]
    print(operation)
    return operation



import weaviate
import os
from openai import AzureOpenAI
import numpy as np
from typing import Dict
import json
import logging

weaviate_client = None

class WeaviateClient:
    def __init__(self, config_path, logger):
        self.config: Dict = None
        self.logger = logger
        if type(config_path) == dict:
            self.config = config_path
        else:
            self.load_config(config_path)
        if "weaviate" not in self.config:
            self.client = None
        else:
            self.config: Dict = self.config["weaviate"]
            self.url = self.config["url"]
            self.api_key = self.config["api_key"]
            self.db = self.config["db"]
            self.limit = self.config["limit"]
            self.client = self.get_client()
        
    def load_config(self, path: str) -> None:
        """
        Load configuration from a specified path.

        :param path: Path to the config file. If an empty path provided,
                     default is `config.json` in the current directory.
        :type path: str
        """
        if path == "":
            current_dir = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(current_dir, "config.json")

        with open(path, "r") as f:
            self.config = json.load(f)

        
    def get_client(self):
        auth_config = weaviate.AuthApiKey(api_key=self.api_key)
        global weaviate_client
        if weaviate_client is None:
            weaviate_client = weaviate.Client(
                url=self.url,
                auth_client_secret=auth_config
            )
        return weaviate_client
    
    def query_bm25(self, properties=["knowledge"], query_string="test", additional="score", limit=3):
        return ((self.client.query
            .get(self.db, properties))
            .with_bm25(
                query=query_string,
                properties=properties  # this does not need to be the same as columns
            )
            .with_additional(additional)
            .with_limit(limit)
            .do()
        )
    
    
    def query_near_text(self, properties=["knowledge"], near_text=["gene"], additional="score", limit=3):
        return ((self.client.query
            .get(self.db, properties))
            .with_near_text({
                "concepts": near_text
            })
            .with_limit(limit)
            .with_additional(additional)
            .do()
        )
    
    
    def query_my_near_text(self, prompt, properties=["knowledge"], additional="score", limit=3):
        vector = {
            "vector": prompt
        }
        return self.query_near_vector(self.db, properties, near_vector=vector, additional=additional, limit=limit)
    
    def query_near_vector(self, properties=["knowledge"], near_vector={}, additional="score", limit=3):
        return ((self.client.query
            .get(self.db, properties))
            .with_near_vector({'vector':near_vector})
            .with_limit(limit)
            .with_additional(additional)
            .do()
        )
    
    def query_with_hybrid(self, properties=["knowledge"], near_vector=[],  near_text='', additional="score", autocut=5):
        return ((self.client.query
            .get(self.db, properties))
            .with_hybrid(query = near_text, vector = near_vector, alpha = 0.8)
            # .with_limit(limit)
            .with_autocut(autocut)
            .with_additional(additional)
            .do()
        )
        
    def object_count(self):
        return ((self.client.query
            .aggregate(self.db)
            .with_meta_count()
            .do()))
        
    
    def get_knowledge(self,embedded_question, max_tokens=4000, max_distance = 0.3, min_score = 0.003, keyword_filter=''):
        if keyword_filter != '':
            knowledge_array = self.query_with_hybrid(near_vector=embedded_question, near_text=keyword_filter, additional=["score"])
        else:
            knowledge_array = self.query_near_vector(near_vector=embedded_question, additional=["distance"], limit=self.limit)
        # knowledge_array = self.query_near_vector(near_vector=embedded_question, additional=["distance"], limit=self.limit)
        knowledge_array = knowledge_array["data"]["Get"][self.config["db"]]
        in_context = []
        distances = []
        cur_tokens = 0
        for knowledge in knowledge_array:
            if keyword_filter != '' and keyword_filter.lower() not in knowledge["knowledge"].lower():
                continue

            if keyword_filter == '' :
                if knowledge['_additional']["distance"] > max_distance:
                    break
            else:
                if float(knowledge['_additional']["score"]) < min_score:
                    break
            
            cur_tokens += len(knowledge["knowledge"].split(" "))
            if cur_tokens > max_tokens:
                break
            in_context.append(knowledge["knowledge"])

            if keyword_filter == '' :
                distances.append(knowledge['_additional']["distance"])
            else:  
                distances.append(float(knowledge['_additional']["score"]))
            
        return in_context, distances



# Copyright (c) 2023 ETH Zurich.
#                    All rights reserved.
#
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
#
# main author: Nils Blach

from abc import ABC, abstractmethod
from typing import List, Dict, Union, Any
import json
import os
import logging


class AbstractLanguageModel(ABC):
    """
    Abstract base class that defines the interface for all language models.
    """

    def __init__(
        self, config_path: Any = "", model_name: str = "", cache: bool = False, logger: logging.Logger = None
    ) -> None:
        """
        Initialize the AbstractLanguageModel instance with configuration, model details, and caching options.

        :param config_path: Path to the config file. Defaults to "".
        :type config_path: str
        :param model_name: Name of the language model. Defaults to "".
        :type model_name: str
        :param cache: Flag to determine whether to cache responses. Defaults to False.
        :type cache: bool
        """
        self.logger = logger
        self.config: Dict = None
        self.model_name: str = model_name
        self.cache = cache
        if self.cache:
            self.response_cache: Dict[str, List[Any]] = {}
        if type(config_path) == dict:
            self.config = config_path
        else:
            self.load_config(config_path)
        self.prompt_tokens: int = 0
        self.completion_tokens: int = 0
        self.cost: float = 0.0

    def load_config(self, path: str) -> None:
        """
        Load configuration from a specified path.

        :param path: Path to the config file. If an empty path provided,
                     default is `config.json` in the current directory.
        :type path: str
        """
        if path == "":
            current_dir = os.path.dirname(os.path.abspath(__file__))
            path = os.path.join(current_dir, "config.json")

        with open(path, "r") as f:
            self.config = json.load(f)

        self.logger.debug(f"Loaded config from {path} for {self.model_name}")

    def clear_cache(self) -> None:
        """
        Clear the response cache.
        """
        self.response_cache.clear()

    @abstractmethod
    def query(self, query: str, num_responses: int = 1) -> Any:
        """
        Abstract method to query the language model.

        :param query: The query to be posed to the language model.
        :type query: str
        :param num_responses: The number of desired responses.
        :type num_responses: int
        :return: The language model's response(s).
        :rtype: Any
        """
        pass

    @abstractmethod
    def get_response_texts(self, query_responses: Union[List[Any], Any]) -> List[str]:
        """
        Abstract method to extract response texts from the language model's response(s).

        :param query_responses: The responses returned from the language model.
        :type query_responses: Union[List[Any], Any]
        :return: List of textual responses.
        :rtype: List[str]
        """
        pass



# Copyright (c) 2023 ETH Zurich.
#                    All rights reserved.
#
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
#
# main author: Nils Blach

from __future__ import annotations
from typing import List

from escargot.operations.operations import Operation


class GraphOfOperations:
    """
    Represents the Graph of Operations, which prescribes the execution plan of thought operations.
    """

    def __init__(self) -> None:
        """
        Initializes a new Graph of Operations instance with empty operations, roots, and leaves.
        The roots are the entry points in the graph with no predecessors.
        The leaves are the exit points in the graph with no successors.
        """
        self.operations: List[Operation] = []
        self.roots: List[Operation] = []
        self.leaves: List[Operation] = []

    def append_operation(self, operation: Operation) -> None:
        """
        Appends an operation to all leaves in the graph and updates the relationships.

        :param operation: The operation to append.
        :type operation: Operation
        """
        self.operations.append(operation)

        if len(self.roots) == 0:
            self.roots = [operation]
        else:
            for leave in self.leaves:
                leave.add_successor(operation)

        self.leaves = [operation]

    def add_operation(self, operation: Operation) -> None:
        """
        Add an operation to the graph considering its predecessors and successors.
        Adjust roots and leaves based on the added operation's position within the graph.

        :param operation: The operation to add.
        :type operation: Operation
        """
        self.operations.append(operation)
        if len(self.roots) == 0:
            self.roots = [operation]
            self.leaves = [operation]
            assert (
                len(operation.predecessors) == 0
            ), "First operation should have no predecessors"
        else:
            if len(operation.predecessors) == 0:
                self.roots.append(operation)
            for predecessor in operation.predecessors:
                if predecessor in self.leaves:
                    self.leaves.remove(predecessor)
            if len(operation.successors) == 0:
                self.leaves.append(operation)



from __future__ import annotations
from typing import Dict, List
import re
import logging
import numpy as np
from escargot.prompter import ESCARGOTPrompter
import ast
import numpy as np

def determine_and_execute(code_snippet, namespace={}):
    local_context = {}
    #detect if there is a print statement in the code
    if 'print' in code_snippet:
        code_snippet = code_snippet.replace('print','')
    try:
        # Try to parse as an expression
        ast.parse(code_snippet, mode='eval')
        # If successful, it's an expression
        return eval(code_snippet, namespace, local_context), 'eval', local_context
    except SyntaxError:
        # add numpy to the namespace
        namespace['np'] = np
        # If there's a syntax error, try to parse as statements
        exec(code_snippet, namespace, local_context)
        return None, 'exec', local_context

class Coder:
    """
    Coder class to manage the code generation and execution.
    """

    def __init__(self) -> None:
        """
        Initialize the Coder instance with the logger.
        """
        self.local_context = {}
        self.step_output = {}
        self.executed_code = {}
        self.instructions = {}

    def execute_code(self, code: str, instruction: str, step_id: str, prompter: ESCARGOTPrompter, logger: logging.Logger) -> str:
        """
        Execute the code and return the output.

        :param code: The code to be executed.
        :type code: str
        :return: The output of the code execution.
        :rtype: str
        """
        tries = 3
        compiled = False

        if len(self.instructions) > 0:
            context = ""
            for cur_step_id, step in self.instructions.items():
                context += "Step " + cur_step_id + ": " + step + "\n"
                context += "Code: " + self.executed_code[cur_step_id] + "\n"
                output = self.step_output[cur_step_id]
                if len(str(output)) > 256:
                    output = str(output)[:256] + "..."
                context += "Local variables: " + str(output) + "\n"

            # code = prompter.adjust_code(code, instruction, context)

        def knowledge_extract(request):
            return prompter.get_knowledge(request,instruction)
        # Add the knowledge_extract function to the local context
        self.local_context["knowledge_extract"] = knowledge_extract
        self.local_context["prompter"] = prompter

        backup_local_context = self.local_context.copy()
        
        while tries > 0 and not compiled:
            try:
                #detect long spaces within the code and remove them, but keep \n
                code = code.replace('            ','')
                if """```python""" in code:
                    code = code.replace("""```python""",'')
                if """```""" in code:
                    code = code.replace("""```""",'')
                if """\n\n""" in code:
                    code = code.replace("""\n\n""",'\n')
                
                result, expression_type, local_context = determine_and_execute(code, self.local_context)
                compiled = True
            except Exception as e:
                logger.warning(f"Could not execute code: {code}. Encountered exception: {e}")
                self.local_context = backup_local_context.copy()
                #debug using the prompter and using the error message
                # prompt = prompter.generate_debug_code_prompt(code, instruction, e)
                # code =prompter.lm.get_response_texts(
                #     prompter.lm.query(prompt, num_responses=1)
                # )[0]
                code = prompter.adjust_code(code, instruction, context)
                tries -= 1
        if compiled:
            self.executed_code[step_id] = code
            self.instructions[step_id] = instruction
            #check for diff between local_context and backup_local_context
            if expression_type == 'eval':
                self.step_output[step_id] = result
            else:
                self.step_output[step_id] = local_context
                self.local_context = self.local_context | local_context
        logger.info(f"Step output: {self.step_output}")

        return code,compiled
    